/******************************************************************************
* 
* CAEN SpA - Front End Division
* Via Vetraia, 11 - 55049 - Viareggio ITALY
* +390594388398 - www.caen.it
*
***************************************************************************//**
* \note TERMS OF USE:
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation. This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. The user relies on the 
* software, documentation and results solely at his own risk.
*
* \file     CAENcom.h
* \brief    CAEN Front End - Communication Layer
* \author   Carlo Tintori (c.tintori@caen.it)
* \version  0.1
* \date     2009/03/23
*
* This library provides the low level functions for the access to the devices.
* By default, this library is based on the CAENVMElib (CAEN VME bridges V1718
* and V2718) and the CAENVMElink (USB and CONET). 
******************************************************************************/

#ifndef __CAENCOM_H
#define __CAENCOM_H


/*###########################################################################*/
/*
** DEFINITIONS
*/
/*###########################################################################*/
/* variable types */
#ifdef WIN32
    #include <windows.h>
    #ifndef int8_t
        #define int8_t  INT8
    #endif
    #ifndef int16_t
        #define int16_t INT16
    #endif
    #ifndef int32_t
        #define int32_t INT32
    #endif
    #ifndef int64_t
        #define int64_t INT64
    #endif
    #ifndef uint8_t
        #define uint8_t  UINT8
    #endif
    #ifndef uint16_t
        #define uint16_t UINT16
    #endif
    #ifndef uint32_t
        #define uint32_t UINT32
    #endif
    #ifndef uint64_t
        #define uint64_t UINT64
    #endif
#else  // LINUX
#endif


/* Link Types */
#define CAENcom_VME    0
#define CAENcom_USB    1
#define CAENcom_CONET  2

/* Error Types */
#define CAENcom_Success             0     /* Operation completed successfully             */
#define CAENcom_VMEBusError        -1     /* VME bus error during the cycle               */
#define CAENcom_CommError          -2     /* Communication error                          */
#define CAENcom_GenericError       -3     /* Unspecified error                            */
#define CAENcom_InvalidParam       -4     /* Invalid parameter                            */
#define CAENcom_InvalidLinkType    -5     /* Invalid Link Type                            */
#define CAENcom_InvalidHandler     -6     /* Invalid device handler                       */
#define CAENcom_MaxDevicesError    -9     /* Maximum number of devices exceeded           */


/*###########################################################################*/
/*
** STRUCTs
*/
/*###########################################################################*/

/**************************************************************************//**
* \struct  CAENcom_DeviceDescr
* \brief   Struct containing the Device access information
*            
*  This structure stores device information for the access
******************************************************************************/
typedef struct {
    int LinkType;       ///< Link Type: VME, CONET, USB, etc... (see defines)
    int LinkNum;        ///< Link Number (index of the USB port or Optical Link controller)
    int ConetNode;      ///< node number in the optical daisy chain (for CONET only)
    int CommHandler;    ///< handler of the device for the communication (VME controller)
    uint32_t VMEBaseAddress; ///< VME Base Address of the board
}
CAENcom_DeviceDescr;



/*##########################################################################*/
/*
** COMMUNICATION LAYER
*/
/*##########################################################################*/

/**************************************************************************//**
* \fn      void CAENcom_DecodeError(int ErrCode, char *ErrMsg)
* \brief   Open the device 
*            
* \param   [IN]  ErrCode: Error code
* \param   [OUT] ErrMsg: string with the error message
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
void CAENcom_DecodeError(int ErrCode, char *ErrMsg);


/**************************************************************************//**
* \fn      int CAENcom_OpenDevice(CAENcom_Handle *handle, dt CAENcom_DeviceType)
* \brief   Open the device 
*            
* \param   [IN]  dd: device descriptor
* \param   [OUT] dev: device handler
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_OpenDevice(CAENcom_DeviceDescr dd, int *dev);


/**************************************************************************//**
* \fn      int CFE_DGTZ_CloseDevice(CFE_DGTZ_Descr *dev)
* \brief   Close the device
*            
* \param   [IN]  dev: device handler
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_CloseDevice(int dev);


/**************************************************************************//**
* \fn      int CAENcom_WriteRegister32(long dev, uint32_t Address, uint32_t *Data)
* \brief   Write a 32 bit register of the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: register address offset 
* \param   [IN]  Data: new register content to write into the device
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_WriteRegister32(int dev, uint32_t Address, uint32_t *Data);


/**************************************************************************//**
* \fn      int CAENcom_WriteRegister16(long dev, uint32_t Address, uint16_t *Data)
* \brief   Write a 16 bit register of the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: register address offset 
* \param   [IN]  Data: new register content to write into the device
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_WriteRegister16(int dev, uint32_t Address, uint16_t *Data);


/**************************************************************************//**
* \fn      int CAENcom_WriteRegister32(long dev, uint32_t Address, uint32_t *Data)
* \brief   Write a 32 bit register of the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: register address offset 
* \param   [IN]  Data: new register content to write into the device
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_ReadRegister32(int dev, uint32_t Address, uint32_t *Data);


/**************************************************************************//**
* \fn      int CAENcom_WriteRegister16(long dev, uint32_t Address, uint16_t *Data)
* \brief   Write a 16 bit register of the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: register address offset 
* \param   [IN]  Data: new register content to write into the device
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_ReadRegister16(int dev, uint32_t Address, uint16_t *Data);


/**************************************************************************//**
* \fn      int CAENcom_BLTRead(int dev, uint32_t Address, uint32_t *Buff, int BltSize, int *nb)
* \brief   Read a block of data from the device using a BLT (32 bit) cycle
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: data space starting address
* \param   [IN]  BltSize: size of the Block Read Cycle (in bytes)
* \param   [OUT] buff: pointer to the data buffer
* \param   [OUT] nw: number of longwords (32 bit) actually read from the device 
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_BLTRead(int dev, uint32_t Address, uint32_t *Buff, int BltSize, int *nw);


/**************************************************************************//**
* \fn      int CAENcom_MBLTRead(int dev, uint32_t Address, uint32_t *Buff, int BltSize, int *nb)
* \brief   Read a block of data from the device using an MBLT (64 bit) cycle
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: data space starting address
* \param   [IN]  BltSize: size of the Block Read Cycle (in bytes)
* \param   [OUT] buff: pointer to the data buffer
* \param   [OUT] nw: number of longwords (32 bit) actually read from the device 
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_MBLTRead(int dev, uint32_t Address, uint32_t *Buff, int BltSize, int *nw);


#endif