/******************************************************************************
* 
* CAEN SpA - Front End Division
* Via Vetraia, 11 - 55049 - Viareggio ITALY
* +390594388398 - www.caen.it
*
***************************************************************************//**
* \note TERMS OF USE:
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation. This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. The user relies on the 
* software, documentation and results solely at his own risk.
*
* \file     DPP.h
* \brief    CAEN Front End - Digital Pulse Processing - Common Header File
* \author   Carlo Tintori (c.tintori@caen.it)
* \version  0.2
* \date     2008/11/03
*
* This is the common header file for the DPPrunner program and the relevant
* libraries and tools
******************************************************************************/
#ifndef __DPP_H
#define __DPP_H


/*###########################################################################*/
/*
** OS portability
*/
/*###########################################################################*/

#ifdef WIN32

    #ifndef _CRT_SECURE_NO_DEPRECATE
        #define _CRT_SECURE_NO_DEPRECATE
    #endif

    #include <time.h>
    #include <sys/timeb.h>
    #include <process.h>
    #include <windows.h>
    #include <winsock.h>

    #define PATH_SEPARATOR '\\'

    // redefine the 'deprecated' functions
    #define popen  _popen  
    #define pclose _pclose 

    #define msleep  Sleep

    #ifndef int8_t
        #define int8_t  INT8
    #endif
    #ifndef int16_t
        #define int16_t INT16
    #endif
    #ifndef int32_t
        #define int32_t INT32
    #endif
    #ifndef int64_t
        #define int64_t INT64
    #endif
    #ifndef uint8_t
        #define uint8_t  UINT8
    #endif
    #ifndef uint16_t
        #define uint16_t UINT16
    #endif
    #ifndef uint32_t
        #define uint32_t UINT32
    #endif
    #ifndef uint64_t
        #define uint64_t UINT64
    #endif

#else  // LINUX

    #include <unistd.h>
    #include <sys/time.h>
    #include <sys/types.h>
    #include <stdint.h>

    #define PATH_SEPARATOR '/'

    #define msleep(t)  usleep((t)*1000)
#endif


// Some Definitions
#define MAX_NB  20          ///< Max. number of boards
#define MAX_NC  (8*MAX_NB)  ///< Max. number of channels

#define NBIT  14                              ///< Number of bit of the digitizer
#define DIGITIZER_TIME_UNIT     10            ///< Sampling period in ns
#define CHANNEL_MEMORY_SIZE     (512*1024)    ///< Memory size (V1724=512 KSample/ch)

#define EHISTO_NBIN   (1<<(NBIT+1))  ///< Num of bins of the energy histogram

#define NUM_TRACES        4  ///< Number of traces in the plot
#define MAX_NUM_PARAMS  100  ///< Max. number of parameters


#define NUM_OUT_FILES     6  ///< Max. number of output files

// output files mask
#define OF_EHISTO   0x01
#define OF_THISTO   0x02
#define OF_ELIST    0x04
#define OF_TLIST    0x08
#define OF_ETLIST   0x10
#define OF_WAVE     0x20

#define MAX_FILE_SIZE   (512*1024*1024)  ///< Max. wave file size (in bytes) 

#define CONFIG_FILE_NAME      "DPPconfig.dpp"      ///< Name of the DPP acquisition config. file
#define DEV_CONFIG_FILE_NAME  "DeviceConfig.dpp"   ///< Name of the hardware config. file

#define WE_MAX_PULSES      1000          ///< Max. number of pulses in a waveform 
#define NOISE_FILE_NAME    "noise.txt"   ///< File name of the noise (emulator)

#define DGTZ_TotNumChannels   (DGTZ->NumBoards * 8)


/**************************************************************************//**
* \struct  DPPparams
* \brief   Parameters for the DPP
******************************************************************************/
typedef struct {
    int k;                ///< Trapezoid rise/fall time (shaping time)
    int m;                ///< Trapezoid flat top
    int M;                ///< Input Signal Decay Time constant
    int a;                ///< Number of samples for the signal average
    int b;                ///< Input Signal Rise time
    int Threshold;        ///< Threshold for the energy trigger
    int FlatTopDelay;     ///< Position in the flat top where to measure the peak 
    int PKavgNs;          ///< Number of samples to average in the peak measurement
    int VoutShift;        ///< Right shift of vout (trapezoid division by k*M)
    int DigitalGain;      ///< Digital gain (must be a power of 2)
    int BLavgNs;          ///< Number of samples for the baseline average
    int Decim;            ///< Decimation (1, 2, 4, 8)
	int DisablePUR;       ///< Disable Pile-Up rejection
	int EnableTrgWin;     ///< Enable Trigger Window
	int TrgWinOffs;       ///< Trigger Window Offset
	int TrgWinWidth;      ///< Trigger Window Width
} DPPparams;


/**************************************************************************//**
* \struct  ACQparams
* \brief   Parameters, variables and memory buffers for the DPP acquisition
******************************************************************************/
typedef struct {

    // Acquisition options
    int RunMode;                    ///< 0 = Full Emulation; 1 = Data from Dgtz, OffLine DPP; 2 = Data from Dgtz, OnLine DPP
    int RecordLength;               ///< Number of samples in the acquisition window
    int CatWindows;                 ///< Concatenate the acq. windows, i.e. assume no dead-time between them (only for RunMode=1)
    int Running;                    ///< 0 = Acq is stopped; 1 = Acq is running
	int SoftStart;                  ///< 0 = Start acq. from Ext Sync; 1 = Start acq. from SW 

    // Scale 
    float BinRescale[MAX_NC];       ///< Rescale factor for the histogram rebinning
    float Bin2EUnit[MAX_NC];        ///< Bin to Energy Unit conversion factor (Energy = BIN * Bin2EUnit + EUnitOffset)
	float EUnitOffset[MAX_NC];      ///< Energy scale offset
    int EUnitMode;                  ///< Energy Unit Mode (0=keep ADC BINs, 1=Use Energy Units)
    char EUnitName[100];            ///< Energy Unit String ("KeV", "MeV", "mV", "fC"...)

    // Plot params
    int PlotMode;                   ///< 0=Oscilloscope 1=Energy Histogram 2=Time Histogram
    int FreeRun;                    ///< If 1, the plot is updated continuously
    int StopPlot;                   ///< For the plot to be stopped
	int RefreshTime;                ///< Waveform Plot Refresh Time
	int VirtualProbe[3];            ///< Virtual probes (DPP internal signal inspection) for Oscilloscope mode
    int TraceEnable;                ///< Trace Enable Mask (4 bits)
    int TraceOffset[NUM_TRACES+1];  ///< Trace Offset (vertical position); index goes from 1 to 4
    int TraceGain[NUM_TRACES+1];    ///< Trace Amplification (vertical scale); index goes from 1 to 4

    // memory buffers for the waveforms
    short *wave;                    ///< Input waveform (from emulator or digitizer)
    short *outwave[NUM_TRACES];     ///< Output waveforms (from the virtual probes)

    // Lists (energies and time tags extracted by the online or offline DPP)
    unsigned short *Elist;          ///< list of peak amplitudes (trapezoid height)
    unsigned long  *Tlist;          ///< list of pulse arrival time (CFD zero crossing)

    // histograms
    int *Ehisto[MAX_NC];            ///< Energy Histogram (15 bit)
    int *Thisto[MAX_NC];            ///< Time tag histogram (time difference between two pulses)
    int *RebinHisto;                ///< Rebinned Energy Histogram (14 bit)
    unsigned long Thmin;            ///< Minimun time for Thisto
    unsigned long Thmax;            ///< Maximum time for Thisto
    int Thbinsize;                  ///< Bin Size of Thisto
    unsigned long Ehmin;            ///< Minimun time for Ehisto
    unsigned long Ehmax;            ///< Maximum time for Ehisto
    int Ehbinsize;                  ///< Bin Size of Ehisto
	int Thnbin;                     ///< Number of Bins in Thisto
    int EhRebin;                    ///< Energy histogram rebinning (1 -> rebin, 0 -> don't rebin)

    // Output Files
    int OutputFileMask;             ///< Bit Mask for enabling output files
    int OutputFileSpFmt;            ///< Spectrum Format: 0=raw data (1 column); 1=raw data (2 column); 2=MCA; 3=SPE
    int OutputFileHead;             ///< 0=Write data only, 1=Include a header before the data (not valid for MCA and SPE formats)
    int OutputFileEU;               ///< 0=write energies in channels; 1=write energies in units (KeV, etc...)
    char OutputFileName[1000];      ///< Output File Name (prefix)
    char OutputDir[1000];           ///< Directory for the output files
    FILE *OutputFile[NUM_OUT_FILES][MAX_NC];     ///< Output file pointers

} ACQparams;


/**************************************************************************//**
* \struct  DGTZparams
* \brief   Parameters, variables and memory buffers for the digitizer
******************************************************************************/
typedef struct {

    // board variables
    int NumBoards;                   ///< Number of digitizers
    int VMEhandler;                  ///< VME bridge handler
    int dev[MAX_NB];                 ///< Device descriptors
    int ChEnableMask[MAX_NB];        ///< Channel Enable Masks

    // channel variables
    int DynamicRange[MAX_NC];        ///< Digitizer input dynamic range (mVpp)
    int InvertInput[MAX_NC];         ///< Invert input signal
    int DCoffset[MAX_NC];            ///< Input DC pffset adjust (common to all channels)
    unsigned long ConfigReg[MAX_NC]; ///< Channel Configuration Register of the digitizer (addr = 0x8000)
    DPPparams DPP[MAX_NC];           ///< DPP algorithm parameters

    // common variables
    int BltSize;                     ///< Size of the block transfer (in bytes)
    int ActiveChannel;               ///< Active Channel (the one plots and measurements refer to)
    int DualTrace;                   ///< DualTrace option
    int AutoSWtrg;                   ///< 0 => single, 1 => continuous
    int AnmonSel;                    ///< Analog Monitor Selection
    int DACout;                      ///< DAC output mode
    int PulserSettings;              ///< Pulser settings (DAC used in pulser mode)
    int AcqMode;                     ///< Acquisition Mode
    int PostTrigger;                 ///< Post Trigger Size (% of the window width)
    int RObufferSize;                ///< Readout Buffer Size (in words)
    unsigned long *buff;             ///< memory buffer for the readout

} DGTZparams;


/**************************************************************************//**
* \struct  WEparams
* \brief   Parameters for the Waveform Emulator
******************************************************************************/
typedef struct {
    int ns;                    ///< Number of samples (waveform length)
    int TimeUnit;              ///< Time Unit (typ. 10 ns)
    int DynamicRange;          ///< Max. Amplitude (FSR) in LSBs
    int BaseLine;              ///< Baseline (DC offset) in LSBs
    int DecayTime;             ///< Decay time constant in TimeUnits
    int RiseTime;              ///< Rise Time in TimeUnits
    int Delay;                 ///< Initial Delay
    int GaussianNoise;         ///< Gaussian Noise Level (in LSB RMS)
    int LowFreqNoise;          ///< Low Frequesncy (50Hz) Noise (base line fluctuation in LSB RMS)
    int NoSaturation;          ///< If 1, discard the pulses that cause saturation
    int RandomMode;            ///< Pulse generation mode: 0=UserDefined, 1=Random
    int PulseAmin;             ///< Minimum pulse amplitude (in LSB)
    int PulseAmax;             ///< Maximum pulse amplitude (in LSB)
    int PulseAstep;            ///< Pulse amplitude Step (in LSB)
    int PulseTmin;             ///< Minimum pulse separation (in Time Units)
    int PulseTmax;             ///< Maximum pulse separation (in Time Units)
    int PulseTstep;            ///< Separation Step (in LSB)
    int np;                    ///< Number of pulses for the User Defined Mode
    int A[WE_MAX_PULSES];      ///< Pulse amplitude vector
    int T[WE_MAX_PULSES];      ///< Pulse time vector
    int UseNoiseFile;          ///< If 1, read a noise signal from a file (noise.txt) and add it recursively to the waveform
} WEparams;

#endif