/******************************************************************************
* 
* CAEN SpA - Front End Division
* Via Vetraia, 11 - 55049 - Viareggio ITALY
* +390594388398 - www.caen.it
*
***************************************************************************//**
* \note TERMS OF USE:
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation. This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. The user relies on the 
* software, documentation and results solely at his own risk.
*
* \file     CAENcom.c
* \brief    CAEN Front End - Communication Layer
* \author   Carlo Tintori (c.tintori@caen.it)
* \version  0.1
* \date     2009/03/23
*
* This library provides the low level functions for the access to the devices.
* By default, this library is based on the CAENVMElib (CAEN VME bridges V1718
* and V2718) and the CAENVMElink (USB and CONET). 
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "../include/CAENcom.h"
#include "../include/CAENVMElib.h"


#define MAX_NUM_DEVICES    100 /* Max. number of devices that can be opened at the same time */

#define VME_SINGLECYC_AM   cvA32_U_DATA  /* VME Address Modifier for single cycles */
#define VME_BLT_AM         cvA32_U_BLT   /* VME Address Modifier for BLT cycles */
#define VME_MBLT_AM        cvA32_U_MBLT  /* VME Address Modifier for MBLT cycles */

/* Global Variables */
int InitDevTable = 1;                          /* Initialize the Dev. table */
CAENcom_DeviceDescr DevTable[MAX_NUM_DEVICES]; /* device descriptor table */
int DevTableUsed[MAX_NUM_DEVICES];             /* used entries in the device descr. table */

/*##########################################################################*/
/*
** COMMUNICATION LAYER
*/
/*##########################################################################*/
#define CAENcom_Success             0     /* Operation completed successfully             */
#define CAENcom_VMEBusError        -1     /* VME bus error during the cycle               */
#define CAENcom_CommError          -2     /* Communication error                          */
#define CAENcom_GenericError       -3     /* Unspecified error                            */
#define CAENcom_InvalidParam       -4     /* Invalid parameter                            */
#define CAENcom_InvalidLinkType    -5     /* Invalid Link Type                            */
#define CAENcom_InvalidHandler     -6     /* Invalid device handler                       */
#define CAENcom_MaxDevicesError    -9     /* Maximum number of devices exceeded           */

/**************************************************************************//**
* \fn      void CAENcom_DecodeError(int ErrCode, char *ErrMsg)
* \brief   Open the device 
*            
* \param   [IN]  ErrCode: Error code
* \param   [OUT] ErrMsg: string with the error message
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
void CAENcom_DecodeError(int ErrCode, char *ErrMsg)
{
    switch (ErrCode) {
        case CAENcom_Success:           strcpy(ErrMsg, "Success"); 
                                        break;
        case CAENcom_VMEBusError:       strcpy(ErrMsg, "VME Bus Error"); 
                                        break;
        case CAENcom_CommError:         strcpy(ErrMsg, "Communication Error"); 
                                        break;
        case CAENcom_GenericError:      strcpy(ErrMsg, "Generic Error"); 
                                        break;
        case CAENcom_InvalidParam:      strcpy(ErrMsg, "Invalid Parameters"); 
                                        break;
        case CAENcom_InvalidLinkType:   strcpy(ErrMsg, "Invalid Link Type"); 
                                        break;
        case CAENcom_InvalidHandler:    strcpy(ErrMsg, "Invalid Devide Handler"); 
                                        break;
        case CAENcom_MaxDevicesError:   strcpy(ErrMsg, "Max. Number of devices exceeded"); 
                                        break;
        default:                        strcpy(ErrMsg, "Unknown Error Code"); 
    }
}


/**************************************************************************//**
* \fn      int CAENcom_OpenDevice(CAENcom_Handle *handle, dt CAENcom_DeviceType)
* \brief   Open the device 
*            
* \param   [IN]  dd: device descriptor
* \param   [OUT] dev: device handler
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_OpenDevice(CAENcom_DeviceDescr dd, int *dev)
{
    int ret, i, pnt;
    long handle;

    if (InitDevTable) {  // First open => set all entries = free
        InitDevTable = 0;
        for(i=0; i<MAX_NUM_DEVICES; i++)
            DevTableUsed[i] = 0;
        pnt = 0;
    } else {  // look for a free entry in the table
        for(i=0; i<MAX_NUM_DEVICES; i++)
            if (DevTableUsed[i] == 0)
                break;
        if (i == MAX_NUM_DEVICES)
            return CAENcom_MaxDevicesError;
        pnt = i;
    }
    DevTableUsed[pnt] = 1;
    memcpy(&DevTable[pnt], &dd, sizeof(CAENcom_DeviceDescr));

    if (dd.LinkType == CAENcom_USB) {
        ret = CAENVME_Init(cvV1718, dd.LinkNum, 0, &handle);
        DevTable[pnt].CommHandler = handle;
    } else if (dd.LinkType == CAENcom_CONET) {
        ret = CAENVME_Init(cvV2718, dd.LinkNum, dd.ConetNode, &handle);
        DevTable[pnt].CommHandler = handle;
    } else if (dd.LinkType == CAENcom_VME) {
        ret = CAENcom_Success;  // HACK: check for CommHandler validity
    } else {
        ret = CAENcom_InvalidLinkType;
    }
    if (ret == cvSuccess)
        *dev = pnt;
    return ret;
}


/**************************************************************************//**
* \fn      int CFE_DGTZ_CloseDevice(CFE_DGTZ_Descr *dev)
* \brief   Close the device
*            
* \param   [IN]  dev: device handler
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_CloseDevice(int dev)
{
    if ((dev < 0) || (dev >= MAX_NUM_DEVICES))
        return CAENcom_InvalidHandler;
    DevTableUsed[dev] = 0;
    if ((DevTable[dev].LinkType == CAENcom_USB) || (DevTable[dev].LinkType == CAENcom_CONET))
        return CAENVME_End(DevTable[dev].CommHandler);
    else 
        return CAENcom_Success;
}


/**************************************************************************//**
* \fn      int CAENcom_WriteRegister32(long dev, uint32_t Address, uint32_t *Data)
* \brief   Write a 32 bit register of the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: register address offset 
* \param   [IN]  Data: new register content to write into the device
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_WriteRegister32(int dev, uint32_t Address, uint32_t *Data)
{
    return CAENVME_WriteCycle(DevTable[dev].CommHandler, DevTable[dev].VMEBaseAddress + Address, Data, VME_SINGLECYC_AM, cvD32);
}

/**************************************************************************//**
* \fn      int CAENcom_WriteRegister16(long dev, uint32_t Address, uint16_t *Data)
* \brief   Write a 16 bit register of the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: register address offset 
* \param   [IN]  Data: new register content to write into the device
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_WriteRegister16(int dev, uint32_t Address, uint16_t *Data)
{
    return CAENVME_WriteCycle(DevTable[dev].CommHandler, DevTable[dev].VMEBaseAddress + Address, Data, VME_SINGLECYC_AM, cvD16);
}


/**************************************************************************//**
* \fn      int CAENcom_WriteRegister32(long dev, uint32_t Address, uint32_t *Data)
* \brief   Write a 32 bit register of the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: register address offset 
* \param   [IN]  Data: new register content to write into the device
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_ReadRegister32(int dev, uint32_t Address, uint32_t *Data)
{
    return CAENVME_ReadCycle(DevTable[dev].CommHandler, DevTable[dev].VMEBaseAddress + Address, Data, VME_SINGLECYC_AM, cvD32);
}

/**************************************************************************//**
* \fn      int CAENcom_WriteRegister16(long dev, uint32_t Address, uint16_t *Data)
* \brief   Write a 16 bit register of the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: register address offset 
* \param   [IN]  Data: new register content to write into the device
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_ReadRegister16(int dev, uint32_t Address, uint16_t *Data)
{
    return CAENVME_ReadCycle(DevTable[dev].CommHandler, DevTable[dev].VMEBaseAddress + Address, Data, VME_SINGLECYC_AM, cvD16);
}


/**************************************************************************//**
* \fn      int CAENcom_BLTRead(int dev, uint32_t Address, uint32_t *Buff, int BltSize, int *nb)
* \brief   Read a block of data from the device using a BLT (32 bit) cycle
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: data space starting address
* \param   [IN]  BltSize: size of the Block Read Cycle (in bytes)
* \param   [OUT] buff: pointer to the data buffer
* \param   [OUT] nw: number of longwords (32 bit) actually read from the device 
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_BLTRead(int dev, uint32_t Address, uint32_t *Buff, int BltSize, int *nw)
{
    int nb, ret;
    ret = CAENVME_FIFOBLTReadCycle(DevTable[dev].CommHandler, DevTable[dev].VMEBaseAddress + Address, Buff, BltSize, VME_BLT_AM, cvD32, &nb);
    *nw = nb/4;
    return ret;
}


/**************************************************************************//**
* \fn      int CAENcom_MBLTRead(int dev, uint32_t Address, uint32_t *Buff, int BltSize, int *nb)
* \brief   Read a block of data from the device using an MBLT (64 bit) cycle
*            
* \param   [IN]  dev: device handler
* \param   [IN]  Address: data space starting address
* \param   [IN]  BltSize: size of the Block Read Cycle (in bytes)
* \param   [OUT] buff: pointer to the data buffer
* \param   [OUT] nw: number of longwords (32 bit) actually read from the device 
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CAENcom_MBLTRead(int dev, uint32_t Address, uint32_t *Buff, int BltSize, int *nw)
{
    int nb, ret;
    ret = CAENVME_FIFOBLTReadCycle(DevTable[dev].CommHandler, DevTable[dev].VMEBaseAddress + Address, Buff, BltSize, VME_MBLT_AM, cvD64, &nb);
    *nw = nb/4;
    return ret;
}
