/******************************************************************************
* 
* CAEN SpA - Front End Division
* Via Vetraia, 11 - 55049 - Viareggio ITALY
* +390594388398 - www.caen.it
*
***************************************************************************//**
* \note TERMS OF USE:
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation. This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. The user relies on the 
* software, documentation and results solely at his own risk.
*
* \file     CFE_DGTZ_lib.c
* \brief    CAEN Front End - Digitizer Library
* \author   Carlo Tintori (c.tintori@caen.it)
* \version  0.1
* \date     2008/11/03
*
* \warning  This library is under development; the list of the functions 
*           is still incomplete
*            
* This library provide functions, structures and definitions for the CAEN
* digitizer family
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "../include/CFE_DGTZ_lib.h"


/*##########################################################################*/
/*
** REGISTER ACCESS LAYER 
*/
/*##########################################################################*/

/**************************************************************************//**
* \fn      int CFE_DGTZ_ResetDevice(CFE_DGTZ_Descr *dev)
* \brief   Reset the device (SW command)
*            
* \param   [IN]  dev: devive handler
* \return  0: success; negative numbers are error codes
******************************************************************************/
int CFE_DGTZ_ResetDevice(int dev)
{
    uint32_t d32 = 0;
    return CAENcom_WriteRegister32(dev, CFE_DGTZ_SWResetReg, &d32);
}


/**************************************************************************//**
* \fn      int CFE_DGTZ_ClearData(CFE_DGTZ_Descr *dev)
* \brief   Clear the data of the acquisition (memory buffer)
*            
* \param   [IN]  dev: devive handler
* \return  0: success; negative numbers are error codes
******************************************************************************/
int CFE_DGTZ_ClearData(int dev)
{
    uint32_t d32 = 0;
    return CAENcom_WriteRegister32(dev, CFE_DGTZ_SWClearReg, &d32);
}


/**************************************************************************//**
* \fn      int CFE_DGTZ_StartAcquisition(CFE_DGTZ_Descr *dev)
* \brief   Start the acquisition
*            
* \param   [IN]  dev: devive handler
* \return  0: success; negative numbers are error codes
******************************************************************************/
int CFE_DGTZ_StartAcquisition(int dev)
{
    uint32_t d32 = 0;
    int ret;

    ret = CAENcom_WriteRegister32(dev, CFE_DGTZ_AcquisitionControlReg, &d32);
    d32 |= 0x00000004;
    ret |= CAENcom_WriteRegister32(dev, CFE_DGTZ_AcquisitionControlReg, &d32);
    return ret;
}


/**************************************************************************//**
* \fn      int CFE_DGTZ_StopAcquisition(CFE_DGTZ_Descr *dev)
* \brief   Stop the acquisition
*            
* \param   [IN]  dev: devive handler
* \return  0: success; negative numbers are error codes
******************************************************************************/
int CFE_DGTZ_StopAcquisition(int dev)
{
    uint32_t d32 = 0;
    int ret;

    ret = CAENcom_WriteRegister32(dev, CFE_DGTZ_AcquisitionControlReg, &d32);
    d32 &= ~0x00000004;
    ret |= CAENcom_WriteRegister32(dev, CFE_DGTZ_AcquisitionControlReg, &d32);
    return ret;
}


/**************************************************************************//**
* \fn      int CFE_DGTZ_SendSWtrig(CFE_DGTZ_Descr *dev)
* \brief   Send a SW trigger
*            
* \param   [IN]  dev: devive handler
* \return  0: success; negative numbers are error codes
******************************************************************************/
int CFE_DGTZ_SendSWtrig(int dev)
{
    uint32_t d32 = 0;
    return CAENcom_WriteRegister32(dev, CFE_DGTZ_SoftTriggerReg, &d32);
}


/**************************************************************************//**
* \fn      int CFE_DGTZ_SendSWsync(CFE_DGTZ_Descr *dev)
* \brief   Send a SW sync (reset time tag)
*            
* \param   [IN]  dev: devive handler
* \return  0: success; negative numbers are error codes
******************************************************************************/
int CFE_DGTZ_SendSWsync(int dev)
{
    uint32_t d32 = 0;
    return CAENcom_WriteRegister32(dev, CFE_DGTZ_SoftSyncReg, &d32);
}


/*##########################################################################*/
/*
** DATA READOUT
*/
/*##########################################################################*/


/**************************************************************************//**
* \fn      int CFE_DGTZ_ReadBlock(int dev, int BltSize, uint32_t *buff, int *nw)
* \brief   Read a block of data from the device
*            
* \param   [IN]  dev: device handler
* \param   [IN]  MaxBlockSize: size of the Block Read Cycle (in bytes)
* \param   [OUT] buff: pointer to the data buffer
* \param   [OUT] nw: number of words read from the device 
* \return  0 = Success; negative numbers are error codes
******************************************************************************/
int CFE_DGTZ_ReadBlock(int dev, int MaxBlockSize, uint32_t *buff, int *nw)
{
    int pnt=0, ret, nw1;

    do {
        ret = CAENcom_MBLTRead(dev, CFE_DGTZ_MultiEventBuffer, buff+pnt, MaxBlockSize, &nw1);
        if ((ret != CAENcom_Success) && (ret != CAENcom_VMEBusError)) 
            return ret;
        pnt += nw1;
    }
    while (ret != CAENcom_VMEBusError);  // read until receive a BERR (end of event)
    *nw = pnt;
    return CAENcom_Success;
}



/**************************************************************************//**
* \fn      int CFE_DGTZ_DPPGetChannelData(uint32_t *buffer, int ch, unit32_t ConfigReg, short *outwave[4], uint16_t *Elist, uint32_t *Tlist, int *np, int *nw)
* \brief   Retrieve the waveforms and/or the pulse energy/timetag of one channel from the event data
*            
* \param   [IN]  buffer: pointer to the first word of the event data
* \param   [IN]  ch: channel
* \param   [IN]  ConfigReg: content of the digitizer Channel Configuration Register
* \param   [OUT] outwave: pointers to the buffers for the waveform
* \param   [OUT] Elist: pointers to the buffers for the energy list
* \param   [OUT] Tlist: pointers to the buffers for the timetag list
* \param   [OUT] np: number of triggers (energy/timetag) found
* \return  Number of event words belonging to that channel
******************************************************************************/
int CFE_DGTZ_DPPGetChannelData(uint32_t *buffer, int ch, uint32_t ConfigReg, int16_t *outwave[4], uint16_t *Elist, uint32_t *Tlist, int *np)
{
    int i, j, pnt=0, TEsearch=0, nw, vp1, vp2; 
    int copied[2];
    int gc1 = 1, gc2 = 1;  // gain correction for trace 1 and 2
    uint32_t *data;

	*np = 0;
    if (!(CFE_DGTZ_CH_MASK(buffer[1]) & (1<<ch)))
        return 0;

    // move the pointer to the requested channel
    if (ConfigReg & CFE_DGTZ_CTRL_INDIV_TRIGGER) {
        int wpnt = 4;  // skip header
        for(i=0; i<ch; i++) {
            if(CFE_DGTZ_CH_MASK(buffer[1]) & (1<<i)) {
                wpnt += buffer[wpnt];
            }
        }
        nw = buffer[wpnt]-1;  // Number of words for that channel
        data = buffer + wpnt + 1;  // pointer to the first word of the requested channel
    } else {
        int nc1 = 0, nc2 = 0;
        // check the channel partecipating mask
        for(i=0; i<8; i++) {
            if(CFE_DGTZ_CH_MASK(buffer[1]) & (1<<i)) {
                nc1++;  // nc1 is the total number of partecipating channels
                if (i < ch)
                    nc2++;  // nc2 is the number of partecipating channels before the active channel
            }
        }
        nw = CFE_DGTZ_NUM_WORD(buffer[0]-4) / nc1;  // Number of words per channel
        data = buffer + nw*nc2 + 4;  // pointer to the first word of the active channel
    }

    vp1 = (ConfigReg >> 12) & 0x3;
    vp2 = (ConfigReg >> 14) & 0x3;
    if (vp1 > 0)
        gc1 = 2;
    if (vp2 > 0)
        gc2 = 2;

    for(i=0; i<nw; i++) {

        copied[0]=0;
        copied[1]=0;

        // Trigger found
        if (data[i] & 0x80000000) {
            if(TEsearch)
                Elist[pnt++] = 0; // Missing energy (pile-up reject)
            Tlist[pnt] = 0;
            TEsearch = 1;
        }

        // data type = time only
        if ((data[i] & 0x80000000) && (ConfigReg & CFE_DGTZ_CTRL_EN_TIME)) {
            Tlist[pnt] = data[i] & 0x3FFFFFFF;
            outwave[3][i*2] = ((data[i] & 1) == 1) * 100;
            outwave[3][i*2+1] = ((data[i] & 1) == 0) * 100;  // o il contrario HACK
            if (i>0) {
                for (j=0; j<3; j++) {
                    if (ConfigReg & CFE_DGTZ_CTRL_DUAL_TRACE) {
                        outwave[j][i*2] = outwave[j][(i-1)*2];  
                        outwave[j][i*2+1] = outwave[j][(i-1)*2+1];
                    } else {
                        outwave[j][i*2] = outwave[j][(i-1)*2+1];  
                        outwave[j][i*2+1] = outwave[j][(i-1)*2+1];
                    }
                }
                copied[0]=1;
                copied[1]=1;
            }

        // data type = energy + 1 sample + 1 digmon
        } else if ((data[i] & 0x40000000) && (ConfigReg & CFE_DGTZ_CTRL_EN_PEAK)) {
            if ((ConfigReg & CFE_DGTZ_CTRL_EN_TIME) && !TEsearch) {
                //printf("\aDATA ERROR (found energy without trigger)!!!\n");  // handle this case properly! HACK
	            // HACK: IF THE FIRST WORD IS AN ENERGY, IT MUST BE DISCARDED (RELEVANT TIMETAG IN THE PREVIOUS EVENT)
				return 0;
            }
            Elist[pnt++] = (unsigned short)(data[i] & 0x7FFF);
            TEsearch = 0;
            outwave[0][i*2]   = (short)((data[i]>>16) & 0x3FFF);
            outwave[0][i*2+1] = (short)((data[i]>>16) & 0x3FFF);
            outwave[2][i*2]   = (short)((data[i]>>15) & 1) * 100;
            outwave[2][i*2+1] = (short)((data[i]>>15) & 1) * 100;
            outwave[3][i*2]   = (short)((data[i]>>31) & 1) * 100;
            outwave[3][i*2+1] = (short)((data[i]>>31) & 1) * 100;
            if (i>0) {
                outwave[1][i*2] = outwave[1][(i-1)*2];
                outwave[1][i*2+1] = outwave[1][(i-1)*2+1];
                copied[1]=1;
            }

        // data type = 2 samples + 2 digmon
        } else {
            if (ConfigReg & CFE_DGTZ_CTRL_DUAL_TRACE) {
                outwave[0][i*2]   = (short)((data[i]>>16) & 0x3FFF);
                outwave[0][i*2+1] = (short)((data[i]>>16) & 0x3FFF);
                outwave[1][i*2]   = (short)(data[i] & 0x3FFF);
                outwave[1][i*2+1] = (short)(data[i] & 0x3FFF);
            } else {
                outwave[0][i*2]   = (short)(data[i] & 0x3FFF);
                outwave[0][i*2+1] = (short)((data[i]>>16) & 0x3FFF);
            }
            outwave[2][i*2]   = (short)((data[i]>>14) & 1) * 100;
            outwave[2][i*2+1] = (short)((data[i]>>15) & 1) * 100;
            outwave[3][i*2]   = (short)((data[i]>>31) & 1) * 100;
            outwave[3][i*2+1] = (short)((data[i]>>31) & 1) * 100;
        }

        for(j=0; j<2; j++) {
            if (!copied[0]) {
                if ((vp1 > 0) && (outwave[0][i*2+j] & 0x2000))
                    outwave[0][i*2+j] |= 0xC000;
                outwave[0][i*2+j] *= gc1;
            }
            if (!copied[1]) {
                if ((vp2 > 0) && (outwave[1][i*2+j] & 0x2000))
                    outwave[1][i*2+j] |= 0xC000;
                outwave[1][i*2+j] *= gc2;
            }
        }
    }
    *np = pnt;
    return nw;
}
