/******************************************************************************
* 
* CAEN SpA - Front End Division
* Via Vetraia, 11 - 55049 - Viareggio ITALY
* +390594388398 - www.caen.it
*
***************************************************************************//**
* \note TERMS OF USE:
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation. This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. The user relies on the 
* software, documentation and results solely at his own risk.
*
* \file     DPPconfig.c
* \brief    CAEN Front End - Digital Pulse Processing - Parameters configuration
* \author   Carlo Tintori (c.tintori@caen.it)
* \version  0.2
* \date     2008/11/03
*
* This library contains the functions for the management of the parameters and
* the programming of the digitizers. It contains also the parser of the 
* configuration files and the management of the socket for the communication
* with the DPP Control Panel (external GUI)
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "../include/CFE_DGTZ_lib.h"
#include "../include/DPP.h"
#include "../include/DPPconfig.h"
#include "../include/keyb.h"
#include "../include/DPPalg.h"
#include "../include/CAENVMElib.h"


#define MAXPENDING 5      /* Maximum outstanding connection requests */

#define MY_SOCK_PORT 5000 /* Port number for DataSocket */


/******************************************************************************
* Command Opodes from the socket
******************************************************************************/
#define OPC_SWTRG       0x01  ///< Send a software trigger
#define OPC_SNGPLOT     0x02  ///< Plot one single event
#define OPC_CLRHISTO    0x04  ///< Clear the histograms
#define OPC_FLUSH       0x08  ///< Flush data
#define OPC_SET         0x20  ///< Set a parameter
#define OPC_SETI        0x30  ///< Set a parameter of an array (index)
#define OPC_RESTART     0x40  ///< Reload config files and restart the acquisition
#define OPC_QUIT        0x80  ///< Quit the program
#define OPC_RUNTOOL     0xFE  ///< Run External Tool
#define OPC_MANCTRL     0xFF  ///< Manual Controller

/**************************************************************************//**
* \struct  ParamTableEntry
* \brief   Parameter Descriptor
******************************************************************************/
typedef struct {
    char Keyword[100];  ///< Parameter name (as reported in the configuration file)
    void *Var;          ///< Pointer to the parameter
    int Type;           ///< Type of the parameter: 0=string, 1=char, 2=short, 4=long, 5=float
    int NumElements;    ///< Number of elements (for arrays)
    int Displacement;   ///< Memory displacement of the array elements
} ParamTableEntry;



// ###########################################################################
// Global Variables
// ###########################################################################
ParamTableEntry ParamTable[MAX_NUM_PARAMS];  ///< Parameter Table
int PTpnt = 0;                               ///< Pointer for the param table

char SockReady = 0;      ///< Socket Ready
char SockBuffer[1000];   ///< Socket Data Buffer
int SocketCommRet = -99; ///< Socket Return code

int FirstInit = 1;       

// ###########################################################################
// Local Functions
// ###########################################################################
// Read an integer from the config file (either decimal or hex)
int ConfigReadNumber(FILE *fp, int *v)
{
    char str[1000];

    fscanf(fp, "%s", str);
    if((str[0]=='0') && (str[1]=='x'))
        return (sscanf(str+2, "%x", v));
    else
        return (sscanf(str, "%d", v));
}

// Add a parameter to the table
void ParamTableAddEntry(char *Keyword, void *Var, int Type, int NumElements, int Displacement, int DefaultValue)
{
    int i;

    if (PTpnt == MAX_NUM_PARAMS) {
        printf("\a\a*** ERROR *** Max. number of parameter exceeded; change MAX_NUM_PARAMS and recompile\n");
        exit(-1);
    }
    strcpy(ParamTable[PTpnt].Keyword, Keyword);
    ParamTable[PTpnt].Var = Var;
    ParamTable[PTpnt].Type = Type;
    ParamTable[PTpnt].NumElements = NumElements;
    ParamTable[PTpnt].Displacement = Displacement;
    for(i=0; i<NumElements; i++) {
        if (Type == 0)  // string
            sprintf((char *)ParamTable[PTpnt].Var + i, "");
        else if (Type == 1)
            *((char *)ParamTable[PTpnt].Var + i) = (char)DefaultValue;
        else if (Type == 2)
            *((short *)ParamTable[PTpnt].Var + i) = (short)DefaultValue;
        else if (Type == 4)
            *((long *)ParamTable[PTpnt].Var + i) = (long)DefaultValue;
        else if (Type == 5)
            *((float *)ParamTable[PTpnt].Var + i) = (float)DefaultValue;
    }
    PTpnt++;
}

// Initialize the parameter table
void InitParamTable(ACQparams *ACQ, WEparams *WE, DGTZparams *DGTZ)
{
    int i,j;

    //                  NAME                 VAR_POINTER          TYPE NE  SZ DEFAULT 
    ParamTableAddEntry("ACQ_RunMode",        &ACQ->RunMode,         4, 1,  0, 0);
    ParamTableAddEntry("ACQ_RecordLength",   &ACQ->RecordLength,    4, 1,  0, 16384);  
    ParamTableAddEntry("ACQ_CatWindows",     &ACQ->CatWindows,      4, 1,  0, 1);      
    ParamTableAddEntry("ACQ_PlotMode",       &ACQ->PlotMode,        4, 1,  0, 0);      
    ParamTableAddEntry("ACQ_FreeRun",        &ACQ->FreeRun,         4, 1,  0, 1);      
    ParamTableAddEntry("ACQ_RefreshTime",    &ACQ->RefreshTime,     4, 1,  0, 200);  
	ParamTableAddEntry("ACQ_VirtualProbe1",  &ACQ->VirtualProbe[0], 4, 1,  0, 3);      
    ParamTableAddEntry("ACQ_VirtualProbe2",  &ACQ->VirtualProbe[1], 4, 1,  0, 0);      
    ParamTableAddEntry("ACQ_VirtualProbe3",  &ACQ->VirtualProbe[2], 4, 1,  0, 0);      
    ParamTableAddEntry("ACQ_TraceEnable",    &ACQ->TraceEnable,     4, 1,  0, 0xF);    
    ParamTableAddEntry("ACQ_TraceOffset",    &ACQ->TraceOffset[0],  4, 5,  sizeof(ACQ->TraceOffset[0]), 0);  
    ParamTableAddEntry("ACQ_TraceGain",      &ACQ->TraceGain[0],    4, 5,  sizeof(ACQ->TraceGain[0]), 1);  
	ParamTableAddEntry("ACQ_Thmin",          &ACQ->Thmin,           4, 1,  0, 0);       
    ParamTableAddEntry("ACQ_Thmax",          &ACQ->Thmax,           4, 1,  0, 16384);   
    ParamTableAddEntry("ACQ_Thbinsize",      &ACQ->Thbinsize,       4, 1,  0, 1);       
    ParamTableAddEntry("ACQ_Ehmin",          &ACQ->Ehmin,           4, 1,  0, 0);       
    ParamTableAddEntry("ACQ_Ehmax",          &ACQ->Ehmax,           4, 1,  0, 16384);   
    ParamTableAddEntry("ACQ_Ehbinsize",      &ACQ->Ehbinsize,       4, 1,  0, 1);       
	ParamTableAddEntry("ACQ_EhRebin",        &ACQ->EhRebin,         4, 1,  0, 1);       
    ParamTableAddEntry("ACQ_EUnitName",      &ACQ->EUnitName,       0, 1,  0, 0);  // string (default value is meaningless)
    ParamTableAddEntry("ACQ_EUnitMode",      &ACQ->EUnitMode,       4, 1,  0, 0); 
    ParamTableAddEntry("ACQ_Bin2EUnit",      &ACQ->Bin2EUnit[0],    5, MAX_NC, sizeof(ACQ->Bin2EUnit[0]), 1);  
    ParamTableAddEntry("ACQ_EUnitOffset",    &ACQ->EUnitOffset[0],  5, MAX_NC, sizeof(ACQ->EUnitOffset[0]), 0);
	
    ParamTableAddEntry("ACQ_OutputFileMask", &ACQ->OutputFileMask,  4, 1,  0, 0);       
    ParamTableAddEntry("ACQ_OutputFileHead", &ACQ->OutputFileHead,  4, 1,  0, 0);
    ParamTableAddEntry("ACQ_OutputFileEU",   &ACQ->OutputFileEU,    4, 1,  0, 0);
    ParamTableAddEntry("ACQ_OutputFileSpFmt",&ACQ->OutputFileSpFmt, 4, 1,  0, 0);
    ParamTableAddEntry("ACQ_OutputDir",      &ACQ->OutputDir,       0, 1,  0, 0);  // string (default value is meaningless)     
    ParamTableAddEntry("ACQ_OutputFileName", &ACQ->OutputFileName,  0, 1,  0, 0);  // string (default value is meaningless)

    ParamTableAddEntry("WE_DynamicRange",    &WE->DynamicRange,     4, 1,  0, 16384);   
    ParamTableAddEntry("WE_DecayTime",       &WE->DecayTime,        4, 1,  0, 5000);    
    ParamTableAddEntry("WE_RiseTime",        &WE->RiseTime,         4, 1,  0, 10);      
    ParamTableAddEntry("WE_Delay",           &WE->Delay,            4, 1,  0, 0);       
    ParamTableAddEntry("WE_BaseLine",        &WE->BaseLine,         4, 1,  0, 300);     
    ParamTableAddEntry("WE_GaussianNoise",   &WE->GaussianNoise,    4, 1,  0, 10);      
    ParamTableAddEntry("WE_LowFreqNoise",    &WE->LowFreqNoise,     4, 1,  0, 10);      
    ParamTableAddEntry("WE_NoSaturation",    &WE->NoSaturation,     4, 1,  0, 1);       
    ParamTableAddEntry("WE_RandomMode",      &WE->RandomMode,       4, 1,  0, 1);       
    ParamTableAddEntry("WE_PulseAmin",       &WE->PulseAmin,        4, 1,  0, 1000);    
    ParamTableAddEntry("WE_PulseAmax",       &WE->PulseAmax,        4, 1,  0, 10000);   
    ParamTableAddEntry("WE_PulseAstep",      &WE->PulseAstep,       4, 1,  0, 1);       
    ParamTableAddEntry("WE_PulseTmin",       &WE->PulseTmin,        4, 1,  0, 1);       
    ParamTableAddEntry("WE_PulseTmax",       &WE->PulseTmax,        4, 1,  0, 10000);   
    ParamTableAddEntry("WE_PulseTstep",      &WE->PulseTstep,       4, 1,  0, 1);       
    ParamTableAddEntry("WE_NumPulses",       &WE->np,               4, 1,  0, 0);      
    ParamTableAddEntry("WE_A",               &WE->A[0],             4, 1,  0, 0);   
    ParamTableAddEntry("WE_T",               &WE->T[0],             4, 1,  0, 0);   
    ParamTableAddEntry("WE_UseNoiseFile",    &WE->UseNoiseFile,     4, 1,  0, 0);        

    ParamTableAddEntry("DPP_M",              &DGTZ->DPP[0].M,            4, MAX_NC, sizeof(DPPparams), 5000);   
    ParamTableAddEntry("DPP_m",              &DGTZ->DPP[0].m,            4, MAX_NC, sizeof(DPPparams), 100);    
    ParamTableAddEntry("DPP_k",              &DGTZ->DPP[0].k,            4, MAX_NC, sizeof(DPPparams), 100);    
    ParamTableAddEntry("DPP_a",              &DGTZ->DPP[0].a,            4, MAX_NC, sizeof(DPPparams), 16);     
    ParamTableAddEntry("DPP_b",              &DGTZ->DPP[0].b,            4, MAX_NC, sizeof(DPPparams), 10);     
    ParamTableAddEntry("DPP_Threshold",      &DGTZ->DPP[0].Threshold,    4, MAX_NC, sizeof(DPPparams), 5000);   
    ParamTableAddEntry("DPP_FlatTopDelay",   &DGTZ->DPP[0].FlatTopDelay, 4, MAX_NC, sizeof(DPPparams), 150);    
    ParamTableAddEntry("DPP_BLavgNs",        &DGTZ->DPP[0].BLavgNs,      4, MAX_NC, sizeof(DPPparams), 2);      
    ParamTableAddEntry("DPP_PKavgNs",        &DGTZ->DPP[0].PKavgNs,      4, MAX_NC, sizeof(DPPparams), 2);      
    ParamTableAddEntry("DPP_DigitalGain",    &DGTZ->DPP[0].DigitalGain,  4, MAX_NC, sizeof(DPPparams), 2);       
    ParamTableAddEntry("DPP_Decim",          &DGTZ->DPP[0].Decim,        4, MAX_NC, sizeof(DPPparams), 1);
    ParamTableAddEntry("DPP_DisablePUR",     &DGTZ->DPP[0].DisablePUR,   4, MAX_NC, sizeof(DPPparams), 0);
    ParamTableAddEntry("DPP_EnableTrgWin",   &DGTZ->DPP[0].EnableTrgWin, 4, MAX_NC, sizeof(DPPparams), 0);
    ParamTableAddEntry("DPP_TrgWinOffs",     &DGTZ->DPP[0].TrgWinOffs,   4, MAX_NC, sizeof(DPPparams), 100);
    ParamTableAddEntry("DPP_TrgWinWidth",    &DGTZ->DPP[0].TrgWinWidth,  4, MAX_NC, sizeof(DPPparams), 64);

    ParamTableAddEntry("DGTZ_ChEnableMask",  &DGTZ->ChEnableMask[0],      4, MAX_NB, sizeof(DGTZ->ChEnableMask[0]), 0xFF);
    ParamTableAddEntry("DGTZ_DynamicRange",  &DGTZ->DynamicRange[0],      4, MAX_NC, sizeof(DGTZ->DynamicRange[0]), 2250);
    ParamTableAddEntry("DGTZ_InvertInput",   &DGTZ->InvertInput[0],       4, MAX_NC, sizeof(DGTZ->InvertInput[0]), 0);
    ParamTableAddEntry("DGTZ_DCoffset",      &DGTZ->DCoffset[0],          4, MAX_NC, sizeof(DGTZ->DCoffset[0]), 50);
    ParamTableAddEntry("DGTZ_BltSize",       &DGTZ->BltSize,              4, 1,  0, -1); 
    ParamTableAddEntry("DGTZ_ActiveChannel", &DGTZ->ActiveChannel,        4, 1,  0, 0);
    ParamTableAddEntry("DGTZ_AnmonSel",      &DGTZ->AnmonSel,             4, 1,  0, 0);
    ParamTableAddEntry("DGTZ_DualTrace",     &DGTZ->DualTrace,            4, 1,  0, 1);
    ParamTableAddEntry("DGTZ_AutoSWtrg",     &DGTZ->AutoSWtrg,            4, 1,  0, 0);
    ParamTableAddEntry("DGTZ_AcqMode",       &DGTZ->AcqMode,              4, 1,  0, 3);
    ParamTableAddEntry("DGTZ_PostTrigger",   &DGTZ->PostTrigger,          4, 1,  0, 80);
    ParamTableAddEntry("DGTZ_DACout",        &DGTZ->DACout,               4, 1,  0, 2);
    ParamTableAddEntry("DGTZ_PulserSettings",&DGTZ->PulserSettings,       4, 1,  0, 0x271013FF);

    DGTZ->NumBoards = 0;
    for(i=0; i<MAX_NB; i++)
        DGTZ->ChEnableMask[i] = 0;

    // initialize the handlers to -1 and all the buffer and file pointers to NULL (meaning uninitialized)
    ACQ->wave = NULL; 
    for(i=0; i<NUM_TRACES; i++)
        ACQ->outwave[i] = NULL;
    ACQ->Elist = NULL;
    ACQ->Tlist = NULL;
    for(i=0; i<MAX_NC; i++) {
        ACQ->Ehisto[i] = NULL;
        ACQ->Thisto[i] = NULL;
    }
    ACQ->RebinHisto = NULL;
    DGTZ->buff = NULL; 
    DGTZ->VMEhandler = -1;
    for(i=0; i<MAX_NB; i++)
        DGTZ->dev[i] = -1;

    for(j=0; j<NUM_OUT_FILES; j++)
        for(i=0; i<MAX_NC; i++)
            ACQ->OutputFile[j][i] = NULL;

    // initialize the variables that didn't take the default setting (arrays, strings and float numbers)
    ACQ->TraceGain[1] = 1;
    ACQ->TraceGain[2] = 1;
    ACQ->TraceGain[3] = 500;
    ACQ->TraceGain[4] = 500;
    ACQ->TraceOffset[1] = 0;
    ACQ->TraceOffset[2] = 0;
    ACQ->TraceOffset[3] = -1000;
    ACQ->TraceOffset[4] = -2000;
    ACQ->TraceEnable = 0xF;
	ACQ->Running = 0;

    sprintf((char *)&ACQ->OutputDir, "OutputFiles");
    sprintf((char *)&ACQ->OutputFileName, "");
    sprintf((char *)&ACQ->EUnitName, "mV");
    for(i=0; i<MAX_NC; i++) 
        ACQ->Bin2EUnit[i] = (float)DGTZ->DynamicRange[i] / (1<<NBIT);
}


// ---------------------------------------------------------------------------
// Allocate and clear the memory buffers 
// ---------------------------------------------------------------------------
int AllocBuffers(ACQparams *ACQ, DGTZparams *DGTZ)
{
    int i, ret, totnb=0;
    int SizeWave, SizeList, SizeEhisto, SizeRebinHisto, SizeThisto, SizeRObuff;

    // calculate the number of bins for Thisto
    for(ACQ->Thnbin=1;;ACQ->Thnbin*=2)  
        if ( (unsigned long)(ACQ->Thnbin) >= ((ACQ->Thmax - ACQ->Thmin) / ACQ->Thbinsize) )
            break;

    SizeWave = ACQ->RecordLength * sizeof(*ACQ->wave);
    SizeList = 1024*256 * sizeof(*ACQ->Elist);
    SizeEhisto = EHISTO_NBIN * sizeof(*ACQ->Ehisto[i]);
    SizeThisto = ACQ->Thnbin * sizeof(*ACQ->Thisto[i]);
    SizeRebinHisto = (1<<NBIT) * sizeof(*ACQ->RebinHisto);
    SizeRObuff = DGTZ->RObufferSize * sizeof(*DGTZ->buff);

    // Waveforms
    ret = 0;
    if ((ACQ->wave = (short *)malloc(SizeWave)) == NULL)
        ret = -1;
    for(i=0; i<NUM_TRACES; i++) {
        if ((ACQ->outwave[i] = (short *)malloc(SizeWave)) == NULL)
            ret = -1;
        memset(ACQ->outwave[i], 0, SizeWave);
    }
    if (ret < 0) { 
        printf("Can't allocate memory buffers for the waveforms\n");
        return(-1);
    }
    totnb += SizeWave * (NUM_TRACES+1);

    // Energy and Time Lists
    ret = 0;
    if ((ACQ->Elist = malloc(SizeList)) == NULL)
        ret = -1;
    if ((ACQ->Tlist = malloc(SizeList)) == NULL)
        ret = -1;
    totnb += SizeList * 2;

    // Energy and Time Histograms
    for(i=0; i<MAX_NC; i++) {
        if ( (DGTZ->ChEnableMask[i/8] & (1 << (i%8))) || ((ACQ->RunMode == 0) && (i==0)) ) {
            if ((ACQ->Ehisto[i] = malloc(SizeEhisto)) == NULL) // Energy Histogram 
                ret = -1;
            if ((ACQ->Thisto[i] = malloc(SizeThisto)) == NULL) // Time Tag Histogram 
                ret = -1;
            totnb += SizeEhisto + SizeThisto;
        }
    }

    if ((ACQ->RebinHisto = malloc(SizeRebinHisto)) == NULL) // Rebinned Histogram 
        ret = -1;
    totnb += SizeRebinHisto;


    if (ret < 0) { 
        printf("Can't allocate memory buffers for the energy/time lists or histograms\n");
        return(-1);
    }
    // clear the histograms 
    ClearHistograms(ACQ);

    // Buffer for the block transfer
    if (ACQ->RunMode > 0) {
        if ((DGTZ->buff = (unsigned long *)malloc(SizeRObuff)) == NULL) {
            printf("Can't allocate memory buffers for the VME readout\n");
            return -1;
        }
    }
    totnb += SizeRObuff;
    printf("> Allocated %d KB of memory\n", totnb/1024);
    return 0;
}

// ---------------------------------------------------------------------------
// Free all buffers already allocated
// ---------------------------------------------------------------------------
void FreeBuffers(ACQparams *ACQ, DGTZparams *DGTZ)
{
    int i;
    if (DGTZ->buff != NULL)
        free(DGTZ->buff);
    if (ACQ->wave != NULL)      
        free(ACQ->wave);
    for(i=0; i<4; i++)
        if (ACQ->outwave[i] != NULL)
            free(ACQ->outwave[i]);
    if (ACQ->Elist != NULL)      
        free(ACQ->Elist);
    if (ACQ->Tlist != NULL)      
        free(ACQ->Tlist);
    for(i=0; i<MAX_NC; i++) {
        if (ACQ->Ehisto[i] != NULL)      
            free(ACQ->Ehisto[i]);
        if (ACQ->Thisto[i] != NULL)      
            free(ACQ->Thisto[i]);
    }
    if (ACQ->RebinHisto != NULL)      
        free(ACQ->RebinHisto);

    ACQ->wave = NULL; 
    for(i=0; i<NUM_TRACES; i++)
        ACQ->outwave[i] = NULL;
    ACQ->Elist = NULL;
    ACQ->Tlist = NULL;
    for(i=0; i<8; i++) {
        ACQ->Ehisto[i] = NULL;
        ACQ->Thisto[i] = NULL;
    }
    DGTZ->buff = NULL; 
}


// ---------------------------------------------------------------------------
// Initialize the digitizer (write the register according to the parameters
// of the acquisition)
// ---------------------------------------------------------------------------
int InitDigitizer(ACQparams *ACQ, DGTZparams *DGTZ)
{
    int b, c, evnw;
    uint32_t chcfg, bforg, custsz, bltne, posttrg, d32;

    // Channel Config. Register
    if (ACQ->RunMode == 2)
        chcfg = CFE_DGTZ_CTRL_INDIV_TRIGGER | (DGTZ->AcqMode << 16) | 
                (DGTZ->DualTrace << 11) | ((ACQ->VirtualProbe[0] & 0x3) << 12) | 
                ((ACQ->VirtualProbe[1] & 0x3) << 14) | ((ACQ->VirtualProbe[2] & 0x7) << 20);
    else if (ACQ->RunMode == 1)
        chcfg = CFE_DGTZ_CTRL_INDIV_TRIGGER | CFE_DGTZ_CTRL_EN_SCOPE;

    // Calculate the buffer organization and the custom size
    for(bforg=10; bforg>=0; bforg--)
        if ( (CHANNEL_MEMORY_SIZE / (1 << bforg)) >= ACQ->RecordLength )
            break;
    custsz = 0;
    if ((CHANNEL_MEMORY_SIZE / (1 << bforg)) > ACQ->RecordLength)
        custsz = ACQ->RecordLength/2;
    // Post Trigger
    posttrg = ACQ->RecordLength * DGTZ->PostTrigger / 200;

    // Calculate the size of the readout buffer (number of long words) and the number of events
    // to transfer with each block transfer
    evnw = (8 * (1 + ACQ->RecordLength/2)) + 4;
    if (DGTZ->BltSize == -1) { // not initialized
        DGTZ->RObufferSize = evnw;
        DGTZ->BltSize = evnw * 4 + 8;  // two words (one VME cycle) more than evnw in order to get berr
        bltne = 1;
    } else {
        if (DGTZ->BltSize < evnw) {
            DGTZ->RObufferSize = evnw;
            bltne = 1;
        } else {
            DGTZ->RObufferSize = DGTZ->BltSize/4;
            bltne = DGTZ->BltSize / (evnw*4);
        }
    }
    if (bltne > 1024)
        bltne = 1024;

    if (!(DGTZ->ChEnableMask[DGTZ->ActiveChannel/8] & (1 << (DGTZ->ActiveChannel%8))))
        printf("\a*** WARNING ***: active channel (%d) is not enabled!\n", DGTZ->ActiveChannel);

    for(b = 0; b < DGTZ->NumBoards; b++) {

		CAENcom_ReadRegister32(DGTZ->dev[b], CFE_DGTZ_AcquisitionControlReg, &d32);
		ACQ->SoftStart = (d32 == 0);

        // event size and readout settings
        CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_BlockOrganizationReg, &bforg);
        CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_CustomSizeReg, &custsz);
        CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_BltEvNumReg, &bltne);

        // Post Trigger
        CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_PostTriggerReg, &posttrg);

        // Write Channel Enable and Trigger Mask
        CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_ChannelEnableMaskReg, &DGTZ->ChEnableMask[b]);

        CAENcom_ReadRegister32(DGTZ->dev[b], CFE_DGTZ_TriggerSourceMaskReg, &d32);
        d32 |= DGTZ->ChEnableMask[b];
        CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_TriggerSourceMaskReg, &d32);
        CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_TriggerOutputMaskReg, &d32);

        // Write DAC-MON settings
        if (DGTZ->DACout == 0) {
            d32 = 5; // Pulser
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_DACoutModeReg, &d32); 
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_DACPulserSettingsReg, &DGTZ->PulserSettings);
        } else {
            d32 = 2; // Analog Inspection
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_DACoutModeReg, &d32); 
            CAENcom_ReadRegister32(DGTZ->dev[b], CFE_DGTZ_AnalogMonitor, &d32); 
            d32 = (d32 & 0xFFFFFF00) | 0x40000000 | (DGTZ->AnmonSel << 4) | DGTZ->ActiveChannel;
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_AnalogMonitor, &d32); 
        }

        // individual channel registers
        for(c = 0; c < 8; c++) {
			uint32_t dgain, blavgns, decim, pkavgns, ftd;

            // Channel Config. Register (read and modify)
            CAENcom_ReadRegister32(DGTZ->dev[b], CFE_DGTZ_ChannelConfigReg, &DGTZ->ConfigReg[b*8+c]);
            DGTZ->ConfigReg[b*8+c] = DGTZ->ConfigReg[b*8+c] & 0x000001FF | chcfg | (DGTZ->InvertInput[b*8+c] << 9);
			DGTZ->ConfigReg[b*8+c] |= 0x400;  // Trigger Windowing
			custsz = (int)(200 / 0.70849729);
			//custsz = (int)(2000 / ACQ->BinRescale[b*8+c]);
			//DGTZ->ConfigReg[b*8+c] |= 0x800000;  // Energy Cut
			//CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_CustomSizeReg, &custsz);
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_ChannelConfigReg, &DGTZ->ConfigReg[b*8+c]);

            // set DC offset (DAC)
            d32 = DGTZ->DCoffset[b*8+c] * 0x7FFF / 50 + 0x8000;
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_InputDCOffsetReg_Ch(c), &d32);  

            // DPP1 registers
            d32 = ((DGTZ->DPP[b*8+c].b + DGTZ->DPP[b*8+c].a) & 0xFF) | ((DGTZ->DPP[b*8+c].a & 0x3F) << 8) | ((DGTZ->DPP[b*8+c].M & 0xFFFF) << 16) |
				  ((DGTZ->DPP[b*8+c].DisablePUR & 1) << 14);
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_DPP1Reg, &d32);

			// DPP2 registers
			switch (DGTZ->DPP[b*8+c].BLavgNs) {
				case 0    : blavgns = 0; 
					        break;
				case 256  : blavgns = 1; 
					        break;
				case 4096 : blavgns = 2; 
					        break;
				default   : blavgns = 3; 
					        break;
			}
            d32 = (DGTZ->DPP[b*8+c].k & 0x3FF) | ((DGTZ->DPP[b*8+c].m & 0x3FF) << 12) | ((DGTZ->DPP[b*8+c].VoutShift & 0x3F) << 24) | (blavgns << 30);  
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_DPP2Reg, &d32);

			// DPP3 registers
			for(pkavgns = 0; pkavgns < 4; pkavgns++)
				if ((1 << (pkavgns*2)) == DGTZ->DPP[b*8+c].PKavgNs)
					break;
			for(decim = 0; decim < 4; decim++)
				if ((1 << decim) == DGTZ->DPP[b*8+c].Decim)
					break;
			for(dgain = 0; dgain < decim; dgain++)
				if ((1 << dgain) == DGTZ->DPP[b*8+c].DigitalGain)
					break;
			if ((DGTZ->DPP[b*8+c].k) > 0)
				ftd = ((DGTZ->DPP[b*8+c].FlatTopDelay + DGTZ->DPP[b*8+c].k + DGTZ->DPP[b*8+c].b) & 0x1FFF);
            else
                ftd = ((DGTZ->DPP[b*8+c].FlatTopDelay) & 0x1FFF);
			d32 = (DGTZ->DPP[b*8+c].Threshold & 0x3FFF) | (decim << 14) | (ftd << 16) | (pkavgns << 28) | (dgain << 30);
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_DPP3Reg, &d32);

            // DPP4 registers
            if (DGTZ->DPP[b*8+c].EnableTrgWin)
                d32 = DGTZ->DPP[b*8+c].TrgWinOffs | ((DGTZ->DPP[b*8+c].TrgWinWidth / 16) << 9);
            else
                d32 = 0;
            CAENcom_WriteRegister32(DGTZ->dev[b], CFE_DGTZ_DPP4Reg, &d32);

        }
    }
    return 0;
}



/*##########################################################################*/
/*
** Public Functions
*/
/*##########################################################################*/


/**************************************************************************//**
* \fn      void ClearHistograms(ACQparams *ACQ)
* \brief   Clear all the histograms
* \param   [IN]  ACQ: struct with the parameters for the acquisition
******************************************************************************/
void ClearHistograms(ACQparams *ACQ)
{
    int i;

    for(i=0; i<MAX_NC; i++) {
        if (ACQ->Ehisto[i] != NULL)
            memset(ACQ->Ehisto[i], 0, EHISTO_NBIN * sizeof(*ACQ->Ehisto[i]));
        if (ACQ->Thisto[i] != NULL)
            memset(ACQ->Thisto[i], 0, ACQ->Thnbin * sizeof(*ACQ->Thisto[i]));
    }
}

/**************************************************************************//**
* \fn      int ParseConfigFile(char *FileName, ACQparams *ACQ, DPPparams *DPP, WEparams *WE, DGTZparams *DGTZ)
* \brief   Read a configuration file and parse the commands
* \param   [IN]  FileName: name of the file to parse
* \param   [IN]  ACQ: struct with the parameters for the acquisition
* \param   [IN]  WE: struct with the parameters for the emulator
* \param   [IN]  DGTZ: struct with the parameters for the digitizers
* \return  0: success, -1: error
******************************************************************************/
int ParseConfigFile(char *FileName, ACQparams *ACQ, WEparams *WE, DGTZparams *DGTZ)
{
    FILE *f_ini;
    int i, j, value, fb, lb;
    float fvalue;
    char Keyword[1000], str[1000], ErrMsg[1000];
    CAENcom_DeviceDescr dd;
    unsigned long addr, d32;
    int ret = 0;

    // Initialize the table that contains all the program parameters (data structures)
    if (PTpnt == 0)
        InitParamTable(ACQ, WE, DGTZ);

    if ( (f_ini = fopen(FileName, "r")) == NULL ) {
        printf("\a*** WARNING *** : Can't open Configuration File %s. Ignored\n", FileName);
        return -2;
    }
    printf("> Reading Configuration File %s\n", FileName);

    while(!feof(f_ini)) {
        fb = 0;
        lb = 0;
        fscanf(f_ini, "%s", Keyword);
        if (Keyword[0] == '#') {  
            fgets(str, 1000, f_ini); // It's a comment => skip the line
            continue;
        }

        // OPEN device
        if (strcmp(Keyword, "OPEN") == 0) {
            fscanf(f_ini, "%s", str);

            // open a VME controller (CAEN bridges V1718/V2718)
            if ((strcmp(str, "V1718") == 0) || (strcmp(str, "V2718") == 0)) {
                ConfigReadNumber(f_ini, &dd.LinkNum);
                if (strcmp(str, "V1718") == 0) {
                    dd.LinkType = CAENcom_USB;
                } else {
                    dd.LinkType = CAENcom_CONET;
                    ConfigReadNumber(f_ini, &dd.ConetNode);
                }
                if ( (ret = CAENcom_OpenDevice(dd, &DGTZ->VMEhandler)) < 0) {
                    CAENcom_DecodeError(ret, ErrMsg), 
                    printf("\a*** WARNING *** : Failure opening %s (%s)\n", str, ErrMsg);
                } else {
                    printf("> %s connected.\n", str);
                }
            // open a digitizer
            } else if (strcmp(str, "DGTZ") == 0) {
                fscanf(f_ini, "%s", str);
                ret = -1;
                if (strcmp(str, "VME") == 0) {
                    ConfigReadNumber(f_ini, &dd.VMEBaseAddress);
                    dd.LinkType = CAENcom_VME;
                    if (DGTZ->VMEhandler < 0) {
                        printf("\a*** WARNING *** : Can't open Digitizer: no VME controller han been opened.\n");
                    } else {
                        dd.CommHandler = DGTZ->VMEhandler;
                        if ( (ret = CAENcom_OpenDevice(dd, &DGTZ->dev[DGTZ->NumBoards])) < 0 ) {
                            CAENcom_DecodeError(ret, ErrMsg), 
                            printf("\a*** WARNING *** : Failure opening %s (%s)\n", str, ErrMsg);
                        }
                    }
                } else if (strcmp(str, "CONET") == 0) {
                    ConfigReadNumber(f_ini, &dd.LinkNum);
                    ConfigReadNumber(f_ini, &dd.ConetNode);
                    ConfigReadNumber(f_ini, &dd.VMEBaseAddress);
                    dd.LinkType = CAENcom_CONET;
                    if ( (ret = CAENcom_OpenDevice(dd, &DGTZ->dev[DGTZ->NumBoards])) < 0 ) {
                        CAENcom_DecodeError(ret, ErrMsg), 
                        printf("\a*** WARNING *** : Failure opening %s (%s)\n", str, ErrMsg);
                    }
                } else {
                    printf("\a*** WARNING *** : unknown DGTZ port (%s). Ignored\n", str);
                }
                if (ret == 0) {
                    printf("> Opened digitizer n. %2d (BaseAddress = %08X).\n", DGTZ->NumBoards, dd.VMEBaseAddress);
                    DGTZ->NumBoards++;
                }

            } else {
                printf("\a*** WARNING *** : unknown device type (%s). Ignored\n", str);
            }
        }

        // WRITE_REGISTER: generic write access to the digitizer
        else if (strncmp(Keyword, "DGTZ_WRITE_REGISTER", 19) == 0) {
            if (Keyword[19] == 'i') {
                ConfigReadNumber(f_ini, &fb);
                if (fb == -1) {  // write all digitizers
                    fb = 0;
                    lb = DGTZ->NumBoards-1;
                } else if (fb < DGTZ->NumBoards) {  // write one digitizer
                    lb = fb;
                } else {  // do nothing
                    fb = 0;
                    lb = -1;  
                }
            } else if (DGTZ->NumBoards > 1) {  // write all digitizers
                fb = 0;
                lb = DGTZ->NumBoards-1;
            }
            ConfigReadNumber(f_ini, &addr);
            ConfigReadNumber(f_ini, &d32);
            for(i=fb; i<=lb; i++) {
                if (DGTZ->dev[i] >= 0) {
                    if (CAENcom_WriteRegister32(DGTZ->dev[i], addr, &d32) < 0) {
                        printf("*** WARNING *** : Digitizer n. %d Write failure (offset=%04X). Ignored.\n", DGTZ->dev[i], addr);
                        ret = -1;
                        continue;
                    }
                }
            }
        }

        // SET: assign a value or a string to a parameter
        else if (strncmp(Keyword, "SET", 3) == 0) {
            fscanf(f_ini, "%s", str);
            if (Keyword[3] == 'i') {
                ConfigReadNumber(f_ini, &fb);
            }
            for(i=0; i<PTpnt; i++){
                if(strcmp(str, ParamTable[i].Keyword) == 0) {
                    if (ParamTable[i].Type == 0) {
                        fscanf(f_ini, "%s", str);
                    } else if (ParamTable[i].Type == 5) {
                        fscanf(f_ini, "%f", &fvalue);
                    } else {
                        ConfigReadNumber(f_ini, &value);
                    }
                    if ((fb == -1) || ((Keyword[3] != 'i') && (ParamTable[i].NumElements > 1))) {  // write all params
                        fb = 0;
                        lb = ParamTable[i].NumElements-1;
                    } else if (fb < ParamTable[i].NumElements) {  // write one only
                        lb = fb;
                    } else {  // do nothing
                        fb = 0;
                        lb = -1;  
                    }
                    for(j=fb; j<=lb; j++) {
                        if (ParamTable[i].Type == 0) {
                            strcpy( ((char *)ParamTable[i].Var + j*ParamTable[i].Displacement), str);
                        } else if (ParamTable[i].Type == 5) {
                            *((float *)((char *)ParamTable[i].Var + j*ParamTable[i].Displacement)) = fvalue;
                        } else {
                            if (ParamTable[i].Type == 1)
                                *((char *)ParamTable[i].Var + j*ParamTable[i].Displacement) = (char)value;
                            else if (ParamTable[i].Type == 2)
                                *((short *)((char *)ParamTable[i].Var + j*ParamTable[i].Displacement)) = (short)value;
                            else if (ParamTable[i].Type == 4)
                                *((long *)((char *)ParamTable[i].Var + j*ParamTable[i].Displacement)) = (long)value;
                        }
                    }
                    break;
                }
            }
            if (i == PTpnt) {
                printf("*** WARNING *** : Invalid Parameter '%s'. Ignored\n", str);
                ret = -3;
                fgets(str, 1000, f_ini);
            }
        }

        // LOAD: parse another config file
        else if (strcmp(Keyword, "LOAD") == 0) {
            fscanf(f_ini, "%s", str);
            if (ParseConfigFile(str, ACQ, WE, DGTZ) < 0)
                return -1;
        }

        else {
            printf("*** WARNING *** : Invalid Command '%s'. Ignored\n", Keyword);
            fgets(str, 1000, f_ini);
            ret = -3;
            continue;
        }
    }
    fclose (f_ini);

    if (ret != 0)
        printf("\a");
    return ret;
}

/**************************************************************************//**
* \fn      int InitAcquisition(ACQparams *ACQ, DPPparams *DPP, WEparams *WE, DGTZparams *DGTZ)
* \brief   Initialize the digitizers and the variables for the acquisition
* \param   [IN]  ACQ: struct with the parameters for the acquisition
* \param   [IN]  WE: struct with the parameters for the emulator
* \param   [IN]  DGTZ: struct with the parameters for the digitizers
* \return  0: success, -1: error
******************************************************************************/
int InitAcquisition(ACQparams *ACQ, WEparams *WE, DGTZparams *DGTZ)
{
    int i,j;
    char str[1000];
    unsigned long d32, c1, c2;
    char OutFileSuffix[NUM_OUT_FILES][4] = {"EH", "TH", "EL", "TL", "TEL"};  

    // Calculate some parameters
    for(i=0; i<MAX_NC; i++) {
        for(DGTZ->DPP[i].VoutShift = 30; DGTZ->DPP[i].VoutShift > 0; DGTZ->DPP[i].VoutShift--)
            if ((1 << DGTZ->DPP[i].VoutShift) <= (DGTZ->DPP[i].M * DGTZ->DPP[i].k))
                break;
        if (DGTZ->DPP[i].k > 0)
            ACQ->BinRescale[i] = (float)(1 << DGTZ->DPP[i].VoutShift) / (float)(DGTZ->DPP[i].M * DGTZ->DPP[i].k);
        else 
            ACQ->BinRescale[i] = 1.0;
    }
    WE->ns = ACQ->RecordLength;
    WE->TimeUnit = DIGITIZER_TIME_UNIT;

    // OffLine DPP Initialization 
    if (ACQ->RunMode < 2)
        DPPalgInit(&DGTZ->DPP[0]);

    if (ACQ->RunMode == 0) {
        WEinit(WE);
        printf("> Run Mode: full emulation\n");
    } else {
        if (FirstInit) {
            FirstInit = 0;
            ParseConfigFile(DEV_CONFIG_FILE_NAME, ACQ, WE, DGTZ);
            if (DGTZ->dev[0] < 0) {
                ACQ->RunMode = 0;
                DPPalgInit(&DGTZ->DPP[0]);
                WEinit(WE);
                printf("> Run Mode: full emulation\n");
            } else {
                printf("> Run Mode: real data from the digitizer\n");
            }
        }
        InitDigitizer(ACQ, DGTZ);
        printf("> Digitizer initialization completed\n");
        // Read Firmware Revision from the digitizer
        CAENcom_ReadRegister32(DGTZ->dev[0], CFE_DGTZ_ChannelFWRevisionReg_Ch(0), &d32);
        printf("> AMC FW Revision:  %ld.%ld (%lx/%ld/%ld)\n", d32>>8 & 0xFF, d32 & 0xFF, d32>>16 & 0xFF, d32>>24 & 0xF, 2000 + (d32>>28 & 0xF));
        // Check the firmware license
        CAENcom_ReadRegister32(DGTZ->dev[0], 0x8158, &c1);
        msleep(100);
        CAENcom_ReadRegister32(DGTZ->dev[0], 0x8158, &c2);
        if (c1 != c2) {
            printf("> WARNING: the firmware is not licensed.\n");
            printf(">          The acquisition will be stopped in %d minutes\n", c2/714);
        }
    }

    // Free and reallocate (or allocate for the first time) the memory buffers
    FreeBuffers(ACQ, DGTZ);
    if (AllocBuffers(ACQ, DGTZ) < 0) 
        return -1;

    // Close and reopen (or open for the first time) the output files
    for(j=0; j<NUM_OUT_FILES; j++)
        for(i=1; i<MAX_NC; i++)
            if (ACQ->OutputFile[j][i] != NULL) {
                fclose(ACQ->OutputFile[j][i]);
                ACQ->OutputFile[j][i] = NULL;
            }

    for(j=0; j<NUM_OUT_FILES; j++) {
        if ((ACQ->OutputFileMask >> j) & 1) {
            char ext[4] = "dat";
            if ((j<2) && ACQ->OutputFileSpFmt == 2)
                strcpy(ext, "mca");
            else if ((j<2) && ACQ->OutputFileSpFmt == 3)
                strcpy(ext, "spe");

            if (ACQ->RunMode == 0) {
                sprintf(str, "%s%c%s_%s.%s", ACQ->OutputDir, PATH_SEPARATOR, ACQ->OutputFileName, OutFileSuffix[j], ext);
                if ((ACQ->OutputFile[j][0] = fopen(str, "w")) == NULL) {
                    printf("\a*** WARNING *** : Can't open output file %s\n", str);
                    printf("                  Data recording will be disabled\n");
                    ACQ->OutputFileMask = 0;
                    return 0;
                }
            } else {
                for(i=0; i<MAX_NC; i++) {
                    sprintf(str, "%s%c%s_%s_%d.%s", ACQ->OutputDir, PATH_SEPARATOR, ACQ->OutputFileName, OutFileSuffix[j], i, ext);
                    if ( (DGTZ->ChEnableMask[i/8] >> (i%8)) & 1) {
                        if ((ACQ->OutputFile[j][i] = fopen(str, "w")) == NULL) {
                            printf("\a*** WARNING *** : Can't open output file %s\n", str);
                            printf("                 Data recording will be disabled\n");
                            ACQ->OutputFileMask = 0;
                            return 0;
                        }
                    }
                }
            }
        }
    }
    return 0;
}

/**************************************************************************//**
* \fn      int EndAcquisition(ACQparams *ACQ, DPPparams *DPP, WEparams *WE, DGTZparams *DGTZ)
* \brief   Free the memory buffers, close all files and handes, atc...
* \param   [IN]  ACQ: struct with the parameters for the acquisition
* \param   [IN]  WE: struct with the parameters for the emulator
* \param   [IN]  DGTZ: struct with the parameters for the digitizers
******************************************************************************/
void EndAcquisition(ACQparams *ACQ, WEparams *WE, DGTZparams *DGTZ)
{
    int i,j;

    FreeBuffers(ACQ, DGTZ);
    WEend();
    for(i=0; i<MAX_NB; i++)
        if (DGTZ->dev[i] > 0) 
            CAENcom_CloseDevice(DGTZ->dev[i]);

    for(j=0; j<NUM_OUT_FILES; j++)
        for(i=1; i<MAX_NC; i++)
            if (ACQ->OutputFile[j][i] != NULL)
                fclose(ACQ->OutputFile[j][i]);
}


/**************************************************************************//**
* \fn      void OpenSocket(void *arg)
* \brief   Open the socket for the communication with the Control Panel
* \param   [IN]  arg: pointer to the arguments (thread)
******************************************************************************/
void OpenSocket(void *arg)
{
#ifdef WIN32
    SOCKET servSock;                 /* Socket descriptor for server */
    SOCKET clntSock;                 /* Socket descriptor for client */
    struct sockaddr_in echoServAddr; /* Local address */
    struct sockaddr_in echoClntAddr; /* Client address */
    unsigned short echoServPort;     /* Server port */
    unsigned int clntLen;            /* Length of client address data structure */
    WSADATA wsaData;                 /* Structure for WinSock setup communication */
    int app;

    echoServPort = MY_SOCK_PORT;     /* first arg:  Local port */

    // Load Winsock 2.0 DLL
    if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0) { 
        SocketCommRet = -1;
        return;
    }

    // Create socket for incoming connections: socket(family,type,protocol)
    if ((servSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) { 
        SocketCommRet = -2;
        return;
    }

    // Construct local address structure 
    memset(&echoServAddr, 0, sizeof(echoServAddr));   /* Zero out structure */
    echoServAddr.sin_family = AF_INET;                /* Internet address family */
    echoServAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    echoServAddr.sin_port = htons(echoServPort);      /* Local port */

    // Bind to the local address
    if (bind(servSock, (struct sockaddr *) &echoServAddr, sizeof(echoServAddr)) < 0) { 
        SocketCommRet = -3;
        return;
    }

    // Mark the socket so it will listen for incoming connections
    if (listen(servSock, MAXPENDING) < 0) { 
        SocketCommRet = -4;
        return;
    }

    SockReady = 0;
    // Run forever
    for (;;) {  
        // Set the size of the in-out parameter
        clntLen = sizeof(echoClntAddr);

        // Wait for a client to connect 
        if ((clntSock = accept(servSock, (struct sockaddr *) &echoClntAddr, &clntLen)) < 0) { 
            SocketCommRet = -5;
            return;
        }
        // clntSock is connected to a client!

        app = recv(clntSock, SockBuffer, 8, 0);
        if (app==4)
            app += recv(clntSock, SockBuffer+4, 4, 0);
        if (app == 8) 
            SockReady = 1;
    }
#endif
}

/**************************************************************************//**
* \fn      int SocketStatus(char *msg)
* \brief   Return the status of the socket
* \param   [OUT] msg: text message that indicates the status
* \return  message code (0=success)
******************************************************************************/
int SocketStatus(char *msg)
{
    sprintf(msg, "");
    if (SocketCommRet == -1)
        sprintf(msg, "\a*** WARNING *** : WSAStartup failed");
    else if (SocketCommRet == -2) 
        sprintf(msg, "\a*** WARNING *** : Socket Create failed");
    else if (SocketCommRet == -3) 
        sprintf(msg, "\a*** WARNING *** : Socket Bind failed");
    else if (SocketCommRet == -4) 
        sprintf(msg, "\a*** WARNING *** : Socket Listen failed");
    else if (SocketCommRet == -5) 
        sprintf(msg, "\a*** WARNING *** : Socket Accept failed");
    else if (SocketCommRet == 0)
        sprintf(msg, "> Socket for the Control Panel successfully opened");
    return SocketCommRet;
}

/**************************************************************************//**
* \fn      int SocketComm(char *Data)
* \brief   Receive a data packet from the socket
* \param   [OUT] Data: data received (array of bytes)
* \return  number of bytes received
******************************************************************************/
int SocketComm(char *Data)
{
    int i, tmp;
    char c = 0;

    // Commands from the socket
    if (SockReady) {
        SockReady = 0;
        for(i=0; i<8; i++)
            Data[i] = SockBuffer[i];
        return 8;
    }

    // Commands from stdin
    if(kbhit()) {
        c = getch();
        for(i=0; i<8; i++)
            Data[i]=0;
        if (c == 'q')
            Data[0] = 0x80;
        if (c == 'd')
            Data[0] = 0xFF;
        if (c == 'c') {
            Data[0] = 0x20;
            Data[2] = 0x01;
            scanf("%d", &tmp);
            Data[7] = (unsigned char)tmp;
        }
		if (c == 'f') 
            Data[0] = OPC_RUNTOOL;

		if (c == 'r') 
            Data[0] = 0x40;

        if ((c >= '1') && (c <= '3')) {
            Data[0] = 0x30;
            Data[2] = 0x09;
            Data[3] = c - '1' + 1;
            Data[7] = -1;
        }

        if (c == 'w') {
            Data[0] = 0x20;
            Data[2] = 0x05;
            Data[7] = 0;
        }
        if (c == 'h') {
            Data[0] = 0x20;
            Data[2] = 0x05;
            Data[7] = 1;
        }
        if (c == 's') {
            Data[0] = 0x20;
            Data[2] = 0x0D;
            Data[7] = 1;
        }
        if (c == 'S') {
            Data[0] = 0x20;
            Data[2] = 0x0D;
            Data[7] = 0;
        }

        return 8;
        } 

    return 0;
}


/**************************************************************************//**
* \fn      int RunTimeCommand(ACQparams *ACQ, WEparams *WE, DGTZparams *DGTZ)
* \brief   Read a command from the socket (or stdin) and execute it
* \param   [IN]  ACQ: struct with the parameters for the acquisition
* \param   [IN]  WE: struct with the parameters for the emulator
* \param   [IN]  DGTZ: struct with the parameters for the digitizers
* \return  Runtime command
******************************************************************************/
int RunTimeCommand(ACQparams *ACQ, WEparams *WE, DGTZparams *DGTZ)
{
    int i, index, param, opcode, value, ret=0;
    char SocketData[1000];  // Data Packet read form the socket
    uint32_t data;
    
    if (SocketComm(SocketData)) {
        index = (unsigned char)SocketData[3];
        param = (unsigned char)SocketData[2];
        opcode = (unsigned char)SocketData[0];
        value = ((SocketData[4]<<24) & 0xFF000000) | 
                ((SocketData[5]<<16) & 0x00FF0000) | 
                ((SocketData[6]<<8)  & 0x0000FF00) | 
                ((SocketData[7])     & 0x000000FF);
        switch(opcode) {
            // QUIT
            case OPC_QUIT :     ret = CMD_QUIT;
                                break;
            // RELOAD CONFIG FILE
            case OPC_RESTART :  ret = CMD_RESTART;
                                break;
            // SW TRIGGER
            case OPC_SWTRG :    if (ACQ->RunMode > 0)
                                    for(i=0; i<DGTZ->NumBoards; i++) 
                                        CFE_DGTZ_SendSWtrig(DGTZ->dev[i]);
                                break;
            // SINGLE PLOT
            case OPC_SNGPLOT :  ACQ->StopPlot = 0;
                                break;
            // CLEAR HISTO
            case OPC_CLRHISTO : ClearHistograms(ACQ);
                                break;
            // FLUSH DATA
            case OPC_FLUSH :    for(i=0; i<DGTZ->NumBoards; i++) 
                                    CFE_DGTZ_ClearData(DGTZ->dev[i]);
                                ret = CMD_CLEAR;
                                break;
            // Manual Controller
            case OPC_MANCTRL :  DigitizerMC(DGTZ);
                                break;

            // Run External Tool
            case OPC_RUNTOOL :  ret = CMD_RUNTOOL;
                                break;

								
								
            // SET (or SETi) PARAM
            case OPC_SET  : index = 0;
            case OPC_SETI : ret = CMD_CLEAR;
                            for(i=0; i<DGTZ->NumBoards; i++)
                                CFE_DGTZ_ClearData(DGTZ->dev[i]);
                            switch (param) {
                                case 0x01 : if ((DGTZ->ChEnableMask[value/8] >> (value%8)) & 1)
                                                DGTZ->ActiveChannel = value;
                                            break;
                                case 0x02 : for(i=0; i<DGTZ_TotNumChannels; i++) {
                                                DGTZ->DCoffset[i] = value;
                                                if (ACQ->RunMode > 0)
                                                    CAENcom_WriteRegister32(DGTZ->dev[i/8], CFE_DGTZ_InputDCOffsetReg_Ch(i%8), &DGTZ->DCoffset[i]);  // BroadCast
                                            }
                                            break;
                                case 0x03 : //ANMON
                                            for(i=0; i<DGTZ->NumBoards; i++) {
                                                data = 2; // Analog Inspection
                                                CAENcom_WriteRegister32(DGTZ->dev[i], CFE_DGTZ_DACoutModeReg, &data); 
                                                CAENcom_WriteRegister32(DGTZ->dev[i], CFE_DGTZ_AnalogMonitor, &data); 
                                                data = (data & 0xFFFFFF00) | 0x40000000 | (DGTZ->AnmonSel << 4) | DGTZ->ActiveChannel;
                                                CAENcom_WriteRegister32(DGTZ->dev[i], CFE_DGTZ_AnalogMonitor, &data); 
                                            }
                                            break;
                                case 0x04 : DGTZ->PostTrigger = value;
                                            data = ACQ->RecordLength * DGTZ->PostTrigger / 200;
                                            if (ACQ->RunMode > 0)
                                                for(i=0; i<DGTZ->NumBoards; i++)
                                                    CAENcom_WriteRegister32(DGTZ->dev[i], CFE_DGTZ_PostTriggerReg, &data);
                                            break;
                                case 0x05 : ret = CMD_INITPLOT;
                                            ACQ->PlotMode = value;
                                            break;
                                case 0x06 : DGTZ->AutoSWtrg = value;
                                            break;
                                case 0x07 : ACQ->FreeRun = value;
                                            ACQ->StopPlot = !value;
                                            break;
                                case 0x08 : ACQ->TraceEnable = value;
                                            break;
                                case 0x09 : if (value == 0xFF)
                                                *(&ACQ->VirtualProbe[0] + index - 1) = *(&ACQ->VirtualProbe[0] + index - 1) + 1;
                                            else 
                                                *(&ACQ->VirtualProbe[0] + index - 1) = value;
                                            ACQ->VirtualProbe[0] &= 0x3;
                                            ACQ->VirtualProbe[1] &= 0x3;
                                            ACQ->VirtualProbe[2] &= 0x7;
                                            for(i=0; i<DGTZ_TotNumChannels; i++) {
                                                DGTZ->ConfigReg[i] = DGTZ->ConfigReg[i] & 0xFF8F0FFF | (ACQ->VirtualProbe[0] << 12) | 
                                                                     (ACQ->VirtualProbe[1] << 14) | (ACQ->VirtualProbe[2] << 20);
                                                if (ACQ->RunMode == 2)
                                                    CAENcom_WriteRegister32(DGTZ->dev[i/8], CFE_DGTZ_ChannelConfigReg, &DGTZ->ConfigReg[i]);
                                                }
                                            break;
                                case 0x0A : ACQ->TraceOffset[index] = value;
                                            break;
                                case 0x0B : ACQ->TraceGain[index] = value;
                                            break;
                                case 0x0D : if (value) {
												for(i=0; i<DGTZ->NumBoards; i++)
													CFE_DGTZ_ClearData(DGTZ->dev[i]);
												ACQ->Running = 1;
												if (ACQ->SoftStart) {
													for(i=0; i<DGTZ->NumBoards; i++)
														CFE_DGTZ_StartAcquisition(DGTZ->dev[i]);
												} else {
													// to do 
												}
                                            } else {
                                                ACQ->Running = 0;
												if (ACQ->SoftStart) {
													for(i=0; i<DGTZ->NumBoards; i++)
														CFE_DGTZ_StopAcquisition(DGTZ->dev[i]);
												} else {
													// to do
												}

											}
                      break;
                  }
                  break;
            }
        }
    return ret;
} 


/**************************************************************************//**
* \fn      int DigitizerMC(DGTZparams *DGTZ)
* \brief   Digitizer Manual Controller (register read/write operations)
* \param   [IN]  DGTZ: struct with the parameters for the digitizers
* \return  0: success, -1: error
******************************************************************************/
int DigitizerMC(DGTZparams *DGTZ)
{
    unsigned long d32, a32, PrintInfo = 1, d=0;
    char c;

    printf("\n\n> Digitizer Manual Controller\n");
    while(1) {
        if (PrintInfo) {
            PrintInfo = 0;
            printf(">> Active Digitizer: %d\n", d);
            // Read Firmware Revision from the digitizer
            if (CAENcom_ReadRegister32(DGTZ->dev[d], CFE_DGTZ_ChannelFWRevisionReg_Ch(0), &d32) < 0) {
                printf("\a*** WARNING *** : Digitizer FW Revision Read failure\n");
                return -1;
            }
            printf(">> AMC FW Revision:                %ld.%ld (%lx/%ld/%ld)\n", d32>>8 & 0xFF, d32 & 0xFF, d32>>16 & 0xFF, d32>>24 & 0xF, 2000 + (d32>>28 & 0xF));
            CAENcom_ReadRegister32(DGTZ->dev[d], CFE_DGTZ_ChannelConfigReg, &d32);
            printf(">> Channel Configuration Register: %08X\n", d32);
            CAENcom_ReadRegister32(DGTZ->dev[d], CFE_DGTZ_ChannelEnableMaskReg, &d32);
            printf(">> Channel Enable Mask:            %08X\n", d32);
            CAENcom_ReadRegister32(DGTZ->dev[d], CFE_DGTZ_BlockOrganizationReg, &d32);
            printf(">> Buffer Organization:            %08X (%d samples)\n", d32, CHANNEL_MEMORY_SIZE/(1<<d32));
            CAENcom_ReadRegister32(DGTZ->dev[d], CFE_DGTZ_DPP1Reg_Ch(0), &d32);
            printf(">> DPP params 1:                   %08X\n", d32);
            CAENcom_ReadRegister32(DGTZ->dev[d], CFE_DGTZ_DPP2Reg_Ch(0), &d32);
            printf(">> DPP params 2:                   %08X\n", d32);
            CAENcom_ReadRegister32(DGTZ->dev[d], CFE_DGTZ_DPP3Reg_Ch(0), &d32);
            printf(">> DPP params 3:                   %08X\n", d32);
        }
        printf(">> Press [r] or [w] to read or write a register, [d] to change digitizer\n");
        c = getch();
        if (c == 'r') {
            printf("   Insert Register Address (offset only): ");
            scanf("%x", &a32);
            CAENcom_ReadRegister32(DGTZ->dev[d], a32, &d32);
            printf(">> Register @ %04X = %08X\n", a32, d32);
        } else if (c == 'w') {
            printf("   Insert Register Address (offset only): ");
            scanf("%x", &a32);
            printf("   Insert Register new content (hex): ");
            scanf("%x", &d32);
            CAENcom_WriteRegister32(DGTZ->dev[d], a32, &d32);
        } else if (c == 'd') {
            if (DGTZ->NumBoards == 1) {
                printf("   There is one digitizer only\n");
            } else {
                printf("   Insert digitizer index (0-%d): ", DGTZ->NumBoards-1);
                scanf("%d", &d);
                printf("\n\n");
                PrintInfo = 1;
            }
        } else {
            break;
        }
    }
    printf(">> Quitting Manual Controller\n\n");
    return 0;
}
