/******************************************************************************
* 
* CAEN SpA - Front End Division
* Via Vetraia, 11 - 55049 - Viareggio ITALY
* +390594388398 - www.caen.it
*
***************************************************************************//**
* \note TERMS OF USE:
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation. This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. The user relies on the 
* software, documentation and results solely at his own risk.
*
* \file     DPPrunner.c
* \brief    CAEN Front End - Digital Pulse Processing - Acquisition Engine
* \author   Carlo Tintori (c.tintori@caen.it)
* \version  1.1
* \date     2009/04/28
* \bug      when acq_mode=6 (T+E), sometime there are two consecutive  
*           identical timetags (maybe it's a FW bug) 
* \todo     Porting to Linux (winexec, socket...)
* \todo     Use a socket for the data transfer to the plotter (gnuplot or others)
*
* The DPPrunner is the Acquisition Engine that manages the peremeters of the
* acquisition, programs the digitizers, reads the data and generates plots and
* output files.
******************************************************************************/

#define DPPrunner_Release        "1.1"
#define DPPrunner_Release_Date   "2009/04/28"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../include/CFE_DGTZ_lib.h"
#include "../include/DPP.h"
#include "../include/DPPconfig.h"
#include "../include/keyb.h"
#include "../include/whplot.h"
#include "../include/DPPalg.h"
#include "../include/CAENVMElib.h"

/******************************************************************************
* Plot and log Refresh times (in ms)
******************************************************************************/
#define HISTO_PLOT_REFRESH_TIME  1000
#define STATISTICS_REFRESH_TIME  1000

#define delta(x)    ((int)(cnt[x]-pcnt[x]))
#define BYTE   0
#define SMPL   1
#define TRIG   2
#define PEAK   3

#define TREF_FREQ   10   // 4.1 ms

// ###########################################################################
// Global Variables
// ###########################################################################
int SuspendAcquisition = 0;  // used for debug


// ###########################################################################
// Functions
// ###########################################################################

// ---------------------------------------------------------------------------
// HistoRebinning: transform the 15 bit histogram obtained from the peak values
// read from the digitizer into a 14 bit histogram applying the bin scaling factor
// BIN(Hout) = BIN(Hin) * bs
// Input Parameters: 
//   Hin: input histogram (32K channels)
//   Hout: output histogram (16K channels)
//   bs: bin scale 
// ---------------------------------------------------------------------------
void HistoRebinning(unsigned long *Hin, unsigned long *Hout, float bs)
{
    int i, j;
    float h;

    j=0;
    h = 0.0;
    for(i=0; i<16384; i++) {
        while((j+1)*bs <= (float)(i+1)) {
            h += (float)Hin[j++];
            if (j >= 32768)
                break;
        }
        h += (float)Hin[j] * (((i+1) - (j*bs)) / bs);
        Hout[i] = (unsigned long)(h+0.5);  // round
        h = (float)Hin[j] * (((j+1)*bs - (i+1)) / bs); // reported to next Hout bin
        j++;
        if (j >= 32768) // end of Hin
            break;
    }
    // fill the rest of the histogram with zeros
    for(j=i; j<16384; j++)
        Hout[j]=0;
}


// ---------------------------------------------------------------------------
// get time in milliseconds
// ---------------------------------------------------------------------------
long get_time()
{
    long time_ms;

#ifdef WIN32
    struct _timeb timebuffer;

    _ftime( &timebuffer );
    time_ms = (long)timebuffer.time * 1000 + (long)timebuffer.millitm;
#else
    struct timeval t1;
    struct timezone tz;

    gettimeofday(&t1, &tz);
    time_ms = (t1.tv_sec) * 1000 + t1.tv_usec / 1000;
#endif
    return time_ms;
}


// ---------------------------------------------------------------------------
// print the rate in Hz, KHz or MHz (t is expressed in ms)
// ---------------------------------------------------------------------------
void PrintRate(int n, int t)
{
    double rate;

    rate = (double)n / t;  // KHz
    if (rate < 1) 
        printf("%6.1f ", rate*1000);
    else if (rate < 1000)
        printf("%6.1fK", rate);
    else 
        printf("%6.1fM", rate/1000);
}



// ---------------------------------------------------------------------------
// Write the header into the relevant output files
// ---------------------------------------------------------------------------
void WriteHeader(ACQparams *ACQ, DGTZparams *DGTZ)
{
    int ch, i; 

    if (ACQ->OutputFileHead) {
        for(ch=0; ch<DGTZ_TotNumChannels; ch++) {
            for(i=2; i<NUM_OUT_FILES; i++) {
                if ( (ACQ->OutputFileMask & (1<<i)) && ( (ACQ->RunMode > 0) && ((DGTZ->ChEnableMask[ch/8] >> (ch%8)) & 1) ||
                                                         (ACQ->RunMode == 0) && (ch == 0) ) ) {
                    fprintf(ACQ->OutputFile[i][ch], "[HEADER]\n");
                    if (ACQ->RunMode == 0)
                        fprintf(ACQ->OutputFile[i][ch], "  Emulation Mode\n");
                    else
                        fprintf(ACQ->OutputFile[i][ch], "  Channel %d\n", ch);
                    if ((i != 5) && (i != 3)) {
                        if (ACQ->OutputFileEU)
                            fprintf(ACQ->OutputFile[i][ch], "  Energy is represented in %s\n", ACQ->EUnitName); 
                        else
                            fprintf(ACQ->OutputFile[i][ch], "  Energy Bin Scale Factor = %f)\n", ACQ->BinRescale);
                    }
                    if ((i==3) || (i==4)) {
                        fprintf(ACQ->OutputFile[i][ch], "  1st column: coarse time in msec\n" );
                        fprintf(ACQ->OutputFile[i][ch], "  2nd column: fine time in nsec\n" );
                    }
                    if (i==4)
                        fprintf(ACQ->OutputFile[i][ch], "  3rd column: energy\n" );
                    fprintf(ACQ->OutputFile[i][ch], "[DATA]\n");
				}
            }
        }
    }
}


// ---------------------------------------------------------------------------
// Save the histograms into the relevant output files
// ---------------------------------------------------------------------------
void SaveHistograms(ACQparams *ACQ, DGTZparams *DGTZ)
{
    int ch, i;

    for(ch=0; ch<DGTZ_TotNumChannels; ch++) {
        if ( ((ACQ->RunMode > 0) && ((DGTZ->ChEnableMask[ch/8] >> (ch%8)) & 1)) || ((ACQ->RunMode == 0) && (ch == 0)) ) {

            // Energy Spectrum
            if (ACQ->OutputFileMask & OF_EHISTO) {
                rewind(ACQ->OutputFile[0][ch]);
                HistoRebinning(ACQ->Ehisto[ch], ACQ->RebinHisto, ACQ->BinRescale[ch]);
                if (ACQ->OutputFileSpFmt == 2) {
                    fprintf(ACQ->OutputFile[0][ch], "<<PMCA SPECTRUM>>\n");
                    fprintf(ACQ->OutputFile[0][ch], "TAG - unknown\n");
                    fprintf(ACQ->OutputFile[0][ch], "DESCRIPTION - Energy Spectrum of Channel %d\n", ch);
                    fprintf(ACQ->OutputFile[0][ch], "GAIN - 6 \n");
                    fprintf(ACQ->OutputFile[0][ch], "THRESHOLD - \n");
                    fprintf(ACQ->OutputFile[0][ch], "LIVE_MODE - 0\n");
                    fprintf(ACQ->OutputFile[0][ch], "PRESET_TIME - 0\n");
                    fprintf(ACQ->OutputFile[0][ch], "LIVE_TIME - \n");
                    fprintf(ACQ->OutputFile[0][ch], "REAL_TIME - \n");
                    fprintf(ACQ->OutputFile[0][ch], "START_TIME - \n");
                    fprintf(ACQ->OutputFile[0][ch], "SERIAL_NUMBER - \n");
                    fprintf(ACQ->OutputFile[0][ch], "<<CALIBRATION>>\n");
                    fprintf(ACQ->OutputFile[0][ch], "LABEL - Energy (%s)\n", ACQ->EUnitName);
                    fprintf(ACQ->OutputFile[0][ch], "0 %f\n", ACQ->EUnitOffset);
                    fprintf(ACQ->OutputFile[0][ch], "1 %f\n", ACQ->Bin2EUnit);
                    fprintf(ACQ->OutputFile[0][ch], "<<DATA>>\n");
                    for (i=0; i<16384; i++)
                        fprintf(ACQ->OutputFile[0][ch], "%d\n", ACQ->RebinHisto[i]);
                    fprintf(ACQ->OutputFile[0][ch], "<<END>>\n");
                } else if (ACQ->OutputFileSpFmt == 3) {
                    fprintf(ACQ->OutputFile[0][ch], "$ENER_FIT:\n");
                    fprintf(ACQ->OutputFile[0][ch], "%f %f\n", ACQ->EUnitOffset, ACQ->Bin2EUnit);
                    fprintf(ACQ->OutputFile[0][ch], "$DATA:\n");
                    fprintf(ACQ->OutputFile[0][ch], "0 16383\n");
                    for (i=0; i<16384; i++)
                        fprintf(ACQ->OutputFile[0][ch], "%d\n", ACQ->RebinHisto[i]);
                } else {
                    for (i=0; i<16384; i++) {
                        if (ACQ->OutputFileSpFmt == 0)
                            fprintf(ACQ->OutputFile[0][ch], "%d\n", ACQ->RebinHisto[i]);
                        else if (ACQ->OutputFileEU == 0)
                            fprintf(ACQ->OutputFile[0][ch], "%8d %d\n", i, ACQ->RebinHisto[i]);
                        else
                            fprintf(ACQ->OutputFile[0][ch], "%15.8f %d\n", (float)i*ACQ->Bin2EUnit[ch]+ACQ->EUnitOffset[ch], ACQ->RebinHisto[i]);
                    }
                }
                fflush(ACQ->OutputFile[0][ch]);
            }

            // Time Spectrum
            if (ACQ->OutputFileMask & OF_THISTO) {
                rewind(ACQ->OutputFile[1][ch]);
                if (ACQ->OutputFileSpFmt == 2) {
                    fprintf(ACQ->OutputFile[1][ch], "MCA file format is not supported for Time Spectra\n");
                } else if (ACQ->OutputFileSpFmt == 3) {
                    fprintf(ACQ->OutputFile[1][ch], "SPE file format is not supported for Time Spectra\n");
                } else {
                    for (i=0; i<ACQ->Thnbin; i++)  
                        fprintf(ACQ->OutputFile[1][ch], "%d\n", ACQ->Thisto[ch][i]);
                }
                fflush(ACQ->OutputFile[1][ch]);
            }
        }
    }
}


// ###########################################################################
// MAIN
// ###########################################################################
int main(int argc, char *argv[])
{
    // Data Structures and pointers
    ACQparams sACQ, *ACQ = &sACQ;     // for the acquisition
    WEparams sWE, *WE = &sWE;         // for the waveform emulator
    DGTZparams sDGTZ, *DGTZ = &sDGTZ; // for the digitizer (access to the hardware devices)

    char ConfigFileName[100] = CONFIG_FILE_NAME;
    char DPPcontrolPanel[100] = "DPPcontrolPanelLV.exe";  // default control panel

    char str[500];
    int i, ch, quit=0, ret, cmd;
    int linecnt = 20;
    int BoardPnt = 0;
    int ActiveChannel;
    int TRefChannel;

    // variables for statistics
    uint64_t cnt[4], pcnt[4];
    uint64_t time[4], ptime[4];
    uint64_t acqtime, histotime, wavetime, stattime;
    uint64_t TimeTag = 0, TimeReference = 0, LiveTime = 0;
    uint64_t TimeTagExtension[MAX_NC];
    uint32_t LastTime[MAX_NC];


    // Readout Buffer pointers and counters
    int ROBpnt=0, ROnw=0, ROevsize=0;
    
    FILE *PlotPipe=NULL;
	
    double mean=0.0, stdev=0.0;
    int nstat=0;


    // ########################################################################

    printf("*******************************************************************\n");
    printf("  CAEN SpA - Front End Division\n");
    printf("  DPPrunner Release %s (%s)\n", DPPrunner_Release, DPPrunner_Release_Date);
    printf("*******************************************************************\n");

    // ************************************************************************
    // System init and config
    // ************************************************************************
    // Read command arguments
    if (argc > 1)
        strcpy(DPPcontrolPanel, argv[1]);

    // Read the configuration file(s) and set parameters
    ParseConfigFile(ConfigFileName, ACQ, WE, DGTZ);
    // Allocate the memory buffers, program the digitizer, initialize the variables, etc...
    if (InitAcquisition(ACQ, WE, DGTZ) < 0)
        goto exit_prog;

    // ************************************************************************
    // Start the Control Panel and the relevant thread and socket
    // ************************************************************************
#ifdef WIN32
    _beginthread(OpenSocket, 0, NULL);
    ret = SocketStatus(str);
    printf("%s\n", str);
    // open the Control Panel
    printf("> Starting the Graphical Control Panel (%s)\n", DPPcontrolPanel);
    WinExec(DPPcontrolPanel, SW_SHOW); 
#else
    // not implemented for Linux
    printf("> Running in console mode (commands from stdin)\n", DPPcontrolPanel);
    printf("\n\nPress 's' to start the acquisition \n");
#endif

    // ************************************************************************
    // Start the plotter (gnuplot in this case)
    // ************************************************************************
    PlotPipe = OpenPlotter();
    SetPlotOptions(PlotPipe, ACQ);

    // ************************************************************************
    // Init the variables
    // ************************************************************************
    WriteHeader(ACQ, DGTZ);
    for(i=0; i<MAX_NC; i++) {
        TimeTagExtension[i]=0;
        LastTime[i]=0;
    }
    for(i=0; i<DGTZ->NumBoards; i++)
        CFE_DGTZ_SendSWsync(DGTZ->dev[i]);
    histotime = get_time();
    wavetime = histotime;
    stattime = histotime;
    for(i=0; i<4; i++) {
        cnt[i]=0;
        pcnt[i]=0;
        ptime[i]=histotime;
    }
    ACQ->StopPlot = 0;

    // ************************************************************************
    // Acquisition Loop
    // ************************************************************************
    while(!quit) {

        // --------------------------------------------------------------------
        // Check for RunTime Commands
        // --------------------------------------------------------------------
        cmd = RunTimeCommand(ACQ, WE, DGTZ);
        if (cmd == CMD_QUIT) {
            quit = 1;
            continue;
        } else if (cmd == CMD_RESTART) {
            // flush data
            ROnw = 0;
            ROBpnt = 0;
            printf("\n\n> Restart the Acquisition\n");
			if ((ACQ->RunMode > 0) && ACQ->Running) {
				if (ACQ->SoftStart) {
					for(i=0; i<DGTZ->NumBoards; i++)
						CFE_DGTZ_StopAcquisition(DGTZ->dev[i]);
				} else {
					// to do
				}
			}
            ParseConfigFile(ConfigFileName, ACQ, WE, DGTZ);
            InitAcquisition(ACQ, WE, DGTZ);
            SetPlotOptions(PlotPipe, ACQ);
            WriteHeader(ACQ, DGTZ);
            for(i=0; i<MAX_NC; i++) {
                TimeTagExtension[i]=0;
                LastTime[i]=0;
            }
			if (ACQ->SoftStart) {
				for(i=0; i<DGTZ->NumBoards; i++) {
					CFE_DGTZ_SendSWsync(DGTZ->dev[i]);
					if ((ACQ->RunMode > 0) && ACQ->Running)
						CFE_DGTZ_StartAcquisition(DGTZ->dev[i]);
				}
			} else {
				// to do
			}
            printf("\n\n");
            linecnt = 20;
        } else if (cmd == CMD_INITPLOT) {
            SetPlotOptions(PlotPipe, ACQ);
        } else if (cmd == CMD_RUNTOOL) {
            FILE *sp;
            sp = fopen("fityk.dat", "w");
            HistoRebinning(ACQ->Ehisto[ActiveChannel], ACQ->RebinHisto, ACQ->BinRescale[ActiveChannel]);
            for (i=0; i<16384; i++)
                fprintf(sp, "%8d %d\n", i, ACQ->RebinHisto[i]);
            fclose(sp);
            WinExec("..\\fityk\\fityk.exe fityk.fit", SW_SHOW);            
		} else {
            SaveHistograms(ACQ, DGTZ);
        }

        if (!ACQ->Running)
            continue;

        // get real time
        acqtime = get_time(); // Time in milliseconds

        // --------------------------------------------------------------------
        // Get an acquisition window (from the digitizer or from the emulator)
        // --------------------------------------------------------------------
        if (ACQ->RunMode == 0) {
            WaveEmulate(ACQ->wave, WE);  // get a waveform from the emulator
            ActiveChannel = 0;
            TRefChannel = 0;
        } else {
            ActiveChannel = DGTZ->ActiveChannel;
            TRefChannel = ActiveChannel;  // can be optionally different !!! HACK
			if ((DGTZ->AutoSWtrg) && (BoardPnt == 0) ) {
				for(i=0; i<DGTZ->NumBoards; i++)
					CFE_DGTZ_SendSWtrig(DGTZ->dev[i]);  // Send a SW trigger
			}

            // If the readout buffer is empty, then read a new block of events from the digitizer
            if (ROBpnt >= ROnw) {
                ROBpnt = 0;
                if ( CFE_DGTZ_ReadBlock(DGTZ->dev[BoardPnt], DGTZ->BltSize, DGTZ->buff, &ROnw) < 0) {
                    printf("\a\a> *** Error *** Block Readout Failed\n");
                    goto exit_prog;
                }
                cnt[BYTE] += ROnw * 4;
                if (ROnw == 0) {  // No data from this board -> read the next one
					BoardPnt = BoardPnt < (DGTZ->NumBoards-1) ? BoardPnt+1 : 0;
                    continue;  // return to the beginning of the while loop
				}
                if ((ACQ->RunMode < 2) && !ACQ->CatWindows)
                    DPPalgInit(&DGTZ->DPP[0]);
            }
        }

        // --------------------------------------------------------------------
        // Analyze the data for all channels and plot the waveforms of the active one
        // --------------------------------------------------------------------
        for(ch=BoardPnt*8; ch<(BoardPnt*8+8); ch++) {
            int np, nw;

            if ((ACQ->RunMode > 0) && (((DGTZ->ChEnableMask[ch/8] >> (ch%8)) & 1) == 0) )  // this channel is not enabled
                continue;

            if (ACQ->RunMode == 0) {
                DPPalg(ACQ->wave, ACQ->RecordLength, ACQ->outwave, ACQ->VirtualProbe, ACQ->Elist, ACQ->Tlist, &np); // apply off-line DPP on the emulated waveform
                cnt[SMPL] += ACQ->RecordLength;
            } else {
                if ((nw = CFE_DGTZ_DPPGetChannelData(DGTZ->buff + ROBpnt, (ch%8), DGTZ->ConfigReg[ch], ACQ->outwave, ACQ->Elist, ACQ->Tlist, &np)) == 0)
                    continue;
                if (DGTZ->ConfigReg[ch] & CFE_DGTZ_CTRL_EN_SCOPE) {
                    cnt[SMPL] += nw*2;
                    if (ch == ActiveChannel)
                        LiveTime += nw * 2 * DIGITIZER_TIME_UNIT;  // Acquisition live time in nsec
                }
                if (ACQ->RunMode == 1)  // apply off-line DPP on the digitizer data
                    DPPalg(ACQ->outwave[0], nw*2, ACQ->outwave, ACQ->VirtualProbe, ACQ->Elist, ACQ->Tlist, &np); 
            }

            // Update histograms and statistics, write output files
            for(i=0; i<np; i++) {
                int deltat;

                // Energy
                if ((ACQ->Elist[i] > 20) && (ACQ->Elist[i] < 16384)) {
                    ACQ->Ehisto[ch][ACQ->Elist[i]]++; // Histogram
                    if (ACQ->OutputFileMask & OF_ELIST)
                        if (ACQ->OutputFileEU)
                            fprintf(ACQ->OutputFile[2][ch], "%6f\n", ACQ->Elist[i] * ACQ->BinRescale[ch] * ACQ->Bin2EUnit[ch] + ACQ->EUnitOffset[ch]);
                        else
                            fprintf(ACQ->OutputFile[2][ch], "%6d\n", ACQ->Elist[i]);
                    if (ch == ActiveChannel) {
                        double e;
                        e = (double)ACQ->Elist[i] * ACQ->BinRescale[ch] * ACQ->Bin2EUnit[ch];
                        //if ((e > 1340) && (e < 1343) && (ACQ->PlotMode == 0))
                        /*if ((ACQ->Elist[i] > 6471) && (ACQ->Elist[i] < 6480) && (ACQ->PlotMode == 0))
                            SuspendAcquisition = 1;*/

                        /*double e;
                        e = (double)ACQ->Elist[i] * ACQ->BinRescale * ACQ->Bin2EUnit;
                        mean += e;
                        stdev += (e*e);
                        nstat++;
                        if (nstat>10000) {
                            mean /= (double)nstat;
                            printf("Mean = %f   StdDev = %f\n", mean, sqrt(stdev/nstat - mean*mean));
                            getch();
                            mean = 0.0;
                            stdev = 0.0;
                            nstat = 0;
                        }*/
                        cnt[PEAK]++;
                        /*if (e < 1500)
                            SuspendAcquisition = 1;*/
                    }
                }

                // Time
                if (ACQ->Tlist[i] < LastTime[ch])  // roll over 
                    TimeTagExtension[ch]++;
                LastTime[ch] = ACQ->Tlist[i];
                TimeTag = ((TimeTagExtension[ch] << 30) | (uint64_t)ACQ->Tlist[i]) * 10;  // in ns

                if (ACQ->OutputFileMask & OF_TLIST) 
                    fprintf(ACQ->OutputFile[3][ch], "%12u %12u\n", (unsigned long)(TimeTag / 1000000), (unsigned long)(TimeTag % 1000000));

                if (ACQ->OutputFileMask & OF_ETLIST) {
                    if (ACQ->OutputFileEU)
                        fprintf(ACQ->OutputFile[4][ch], "%12u %12u %6f\n", (unsigned long)(TimeTag / 1000000), (unsigned long)(TimeTag % 1000000), 
                            ACQ->Elist[i] * ACQ->BinRescale[ch] * ACQ->Bin2EUnit[ch] + ACQ->EUnitOffset[ch]);
                    else
                        fprintf(ACQ->OutputFile[4][ch], "%12u %12u %6d\n", (unsigned long)(TimeTag / 1000000), (unsigned long)(TimeTag % 1000000), ACQ->Elist[i]);
                }

                deltat = (int)(TimeTag - TimeReference) / ACQ->Thbinsize;
                if ((deltat & 0x7FFFFFFF) < ACQ->Thnbin) {
                    ACQ->Thisto[ch][deltat]++;
                }
                if (ch == TRefChannel)
                    TimeReference = TimeTag;
                if (ch == ActiveChannel) {
                    cnt[TRIG]++;
                    /*if (deltat < 5)
                        SuspendAcquisition = 1;*/
                }
            }

			/* Write the waveform to the relevant output files */
			if (ACQ->OutputFileMask & OF_WAVE) {
				fprintf(ACQ->OutputFile[5][ch], "StartWave : %d samples\n", ACQ->RecordLength);
				if (ftell(ACQ->OutputFile[5][ch]) < MAX_FILE_SIZE)
					for(i=0; i<ACQ->RecordLength; i++) 
						fprintf(ACQ->OutputFile[5][ch], "%d\n", ACQ->outwave[0][i]);
			}

            if ((acqtime-wavetime) > ACQ->RefreshTime) {
                // Plot the waveforms of the active channel
                if ( (((ACQ->PlotMode == 0) || (ACQ->PlotMode == 3)) && (ch == ActiveChannel)) &&
                     ((ACQ->RunMode == 0) || (DGTZ->ConfigReg[ActiveChannel] & CFE_DGTZ_CTRL_EN_SCOPE)) && 
                     (!ACQ->StopPlot || SuspendAcquisition)) {

                    if (ACQ->PlotMode == 0)
						PlotWaveforms(PlotPipe, ACQ, ActiveChannel);
					else
						PlotFFT(PlotPipe, ACQ, ActiveChannel);
                    wavetime = acqtime;
                    // clear the memory buffers of the digitizer
					if (ACQ->RunMode > 0) {
						for(i=0; i<DGTZ->NumBoards; i++)
							CFE_DGTZ_ClearData(DGTZ->dev[i]);
					}
                    if (!ACQ->FreeRun)
                        ACQ->StopPlot = 1;
                    if (SuspendAcquisition) {
                        printf(">>> Readout Suspended. Press a key to continue...\n");
                        SuspendAcquisition = 0;
                        getch();
                    }
                }
            }

            if (ACQ->RunMode == 0)  // in emulation mode, there is only one virtual channel
                break;
        }

        // move the pointer to next event in the data buffer
        if (ACQ->RunMode > 0) {
            ROBpnt += CFE_DGTZ_NUM_WORD(DGTZ->buff[ROBpnt]);
			// in case of MBLT on the VME, there may be one extra fake word due to the 64 bit transfer
			if ((ROnw - ROBpnt) == 1)
				ROBpnt = ROnw;
        }


        // --------------------------------------------------------------------
        // Calculate and print Statistics
        // --------------------------------------------------------------------
        if ((acqtime-stattime) > STATISTICS_REFRESH_TIME) {
            if (linecnt++ == 20) {
                printf("\n ---------- Rates ----------- | ---------- Statistics ---------- \n");
                printf(  "  byte   smpl   trg    peak   |  Ntrg    Npk    PUR   LiveTime   \n");
                printf(  " ---------------------------- | -------------------------------- \n");
                linecnt = 0;
            }
				
            for(i=0; i<4; i++)
                PrintRate(delta(i), (int)(acqtime-histotime));
            printf("  | %5d", delta(TRIG));
            printf("  %5d", delta(PEAK));
            if (delta(TRIG) > 0) 
                printf(" %5d%%", 100-delta(PEAK)*100/delta(TRIG));
            else
                printf(" -    ");
            if ((ACQ->RunMode > 0) && (DGTZ->ConfigReg[ActiveChannel] & CFE_DGTZ_CTRL_EN_SCOPE))
                printf("    %6.2f%%\n", (float)LiveTime / 10000000);
            else
                printf("    %6.2f%%\n", 100.0);

            // Update counters
            LiveTime = 0;
            for(i=0; i<4; i++) {
                pcnt[i] = cnt[i];
                ptime[i] = time[i];
            }
            stattime = acqtime;
        }

        // --------------------------------------------------------------------
        // plot/save the histograms (executed every second) 
        // --------------------------------------------------------------------
        if (((ACQ->PlotMode == 1) || (ACQ->PlotMode == 2)) && (acqtime-histotime) > HISTO_PLOT_REFRESH_TIME) {
            // Plot the histogram
            if (!ACQ->StopPlot) {
                if (ACQ->PlotMode == 1)
                    HistoRebinning(ACQ->Ehisto[ActiveChannel], ACQ->RebinHisto, ACQ->BinRescale[ActiveChannel]);
                PlotHistograms(PlotPipe, ACQ, ActiveChannel);
                if (!ACQ->FreeRun)
                    ACQ->StopPlot = 1;
            }
            histotime = acqtime;
        }

    }

// ************************************************************************
// Close everything
// ************************************************************************
exit_prog:
	//fclose(stream);
    SaveHistograms(ACQ, DGTZ);
    EndAcquisition(ACQ, WE, DGTZ);
    ClosePlotter(PlotPipe);
}



