/******************************************************************************
* 
* CAEN SpA - Front End Division
* Via Vetraia, 11 - 55049 - Viareggio ITALY
* +390594388398 - www.caen.it
*
***************************************************************************//**
* \note TERMS OF USE:
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation. This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. The user relies on the 
* software, documentation and results solely at his own risk.
*
* \file     whplot.c
* \brief    CAEN Front End - Waveform and Histogram plotter
* \author   Carlo Tintori (c.tintori@caen.it)
* \version  0.1
* \date     2009/03/05
*
* whplot is a library that allows the DPPrunner to plot waveforms and 
* histograms. It saves the data read from the digitizer (or from the 
* emulator) into a text file and send the plotiing commands to gnuplot through
* a pipe. Thus, whplot is just a simple server that transfer the data and 
* the commands from DPPrunner to gnuplot, which is the actual plotting tool.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../include/DPP.h"
#include "../include/fft.h"

/******************************************************************************
* Executable file for 'gnuplot'
* NOTE: use pgnuplot instead of wgnuplot under Windows, otherwise the pipe 
*       will not work
******************************************************************************/
#ifdef WIN32
    #define GNUPLOT_COMMAND  "pgnuplot"
#else
    #define GNUPLOT_COMMAND  "gnuplot"
#endif

static char WaveLabel[3][8][30] = { "Input", "Delta", "CFD", "Trapezoid", "", "", "", "",
                                    "Input", "S3", "Digital Combo", "Trapezoid Baseline", "", "", "", "",
                                    "Trigger", "Armed", "PkRun", "PkAbort", "Peaking", "PkHolfOff", "Flat", "Zero"};
                       

/**************************************************************************//**
* \fn      FILE *OpenPlotter(void)
* \brief   Open the plotter (i.e. open gnuplot with a pipe)
* \return  file pointer (pipe); return NULL in case of error
******************************************************************************/
FILE *OpenPlotter(void)
{
    FILE *gnuplot;

    gnuplot = popen(GNUPLOT_COMMAND, "w"); // open the pipe
    //system(GNUPLOT_COMMAND);
    return gnuplot;
}


/**************************************************************************//**
* \fn      void SetPlotOptions(FILE *gnuplot, ACQparams *ACQ)
* \brief   Set the plotting options (labels, X-Y scale, etc...)
* \param   [IN]  gnuplot: gnuplot pipe
* \param   [IN]  ACQ: struct with the parameters for the acquisition
******************************************************************************/
void SetPlotOptions(FILE *gnuplot, ACQparams *ACQ)
{
    if (ACQ->PlotMode == 0) {
        fprintf(gnuplot, "set xlabel \"Time (1 sample = %dns)\"\n", DIGITIZER_TIME_UNIT);
        fprintf(gnuplot, "load 'plot_wave.cfg'\n");
    } else if (ACQ->PlotMode == 1) {
        fprintf(gnuplot, "set ylabel \"Counts\"\n");
        fprintf(gnuplot, "load 'plot_histo.cfg'\n");
    } else if (ACQ->PlotMode == 2) {
        fprintf(gnuplot, "set xlabel \"1 BIN = %dns\"\n", DIGITIZER_TIME_UNIT);
        fprintf(gnuplot, "set ylabel \"Counts\"\n");
        fprintf(gnuplot, "load 'plot_histo.cfg'\n");
    } else if (ACQ->PlotMode == 3) {
        fprintf(gnuplot, "load 'plot_fft.cfg'\n");
    } 

	fflush(gnuplot);
}

/**************************************************************************//**
* \fn      void PlotWaveforms(FILE *gnuplot, ACQparams *ACQ, int Channel)
* \brief   Plot the waveforms
* \param   [IN]  gnuplot: gnuplot pipe
* \param   [IN]  ACQ: struct with the parameters for the acquisition
* \param   [IN]  Channel: channel to plot
******************************************************************************/
void PlotWaveforms(FILE *gnuplot, ACQparams *ACQ, int Channel)
{
    int p, i, comma=0;
    char tmp[500], cmd[500];
    FILE *plotfile;

    if (ACQ->RunMode == 0)
        fprintf(gnuplot, "set title \"Emulation Mode\"\n");    
    else
        fprintf(gnuplot, "set title \"Channel %d\"\n", Channel);
    fprintf(gnuplot, "set ylabel \"1 BIN = %6.3f %s\"\n", ACQ->Bin2EUnit[Channel], ACQ->EUnitName);
    // str contains the plot command for gnuplot
    sprintf(cmd, "plot ");
    for(p=1; p<=NUM_TRACES; p++) {
        if (ACQ->TraceEnable & (1<<(p-1))) {
            sprintf(tmp, "trace%d.txt", p);
            plotfile = fopen(tmp, "w");
            // Write plot data to file
            for(i=0; i<ACQ->RecordLength; i++)
                fprintf(plotfile, "%d\n", ACQ->outwave[p-1][i] * ACQ->TraceGain[p] + ACQ->TraceOffset[p]);
            fclose(plotfile);
            if (comma)
                strcat(cmd, ", ");
            if (p == 4)
                sprintf(tmp, "'trace%d.txt' title '%s' with lines %d", p, "Trigger", p);
            else
                sprintf(tmp, "'trace%d.txt' title '%s' with lines %d", p, WaveLabel[p-1][ACQ->VirtualProbe[p-1]], p);
            strcat(cmd, tmp);
            comma = 1;
        }
    }
    fprintf(gnuplot, "%s\n", cmd); 
    fflush(gnuplot);
}

/**************************************************************************//**
* \fn      void PlotFFT(FILE *gnuplot, ACQparams *ACQ, int Channel)
* \brief   Plot the FFT of the virtual probe 1
* \param   [IN]  gnuplot: gnuplot pipe
* \param   [IN]  ACQ: struct with the parameters for the acquisition
* \param   [IN]  Channel: channel to plot
******************************************************************************/
void PlotFFT(FILE *gnuplot, ACQparams *ACQ, int Channel)
{
    int i, FFTns;
	double *FFTout;
    FILE *plotfile;

	FFTout = malloc(ACQ->RecordLength * sizeof(double));
	if(FFTout == NULL)
		return;
    if (ACQ->RunMode == 0)
        fprintf(gnuplot, "set title \"Emulation Mode\"\n");    
    else
        fprintf(gnuplot, "set title \"Channel %d\"\n", Channel);
	FFTns = FFT(ACQ->outwave[0], FFTout, ACQ->RecordLength, HANNING_FFT_WINDOW);
    // str contains the plot command for gnuplot
    plotfile = fopen("fft.txt", "w");
    // Write plot data to file
    for(i=0; i<FFTns; i++)
        fprintf(plotfile, "%f %f\n", i * 100.0 / (2*FFTns), FFTout[i]);
    fclose(plotfile);
    fprintf(gnuplot, "plot 'fft.txt' title 'FFT' with line 1\n");  
    fflush(gnuplot);
	free(FFTout);
}

/**************************************************************************//**
* \fn      void PlotHistograms(FILE *gnuplot, ACQparams *ACQ, int Channel)
* \brief   Plot the histogram
* \param   [IN]  gnuplot: gnuplot pipe
* \param   [IN]  ACQ: struct with the parameters for the acquisition
* \param   [IN]  Channel: channel to plot
******************************************************************************/
void PlotHistograms(FILE *gnuplot, ACQparams *ACQ, int Channel)
{
    int i;
    uint64_t HistoEntries = 0;
    FILE *plotfile;

    if (ACQ->PlotMode > 0) {

        if (ACQ->PlotMode == 1) {
            if (ACQ->EUnitMode == 0)
                fprintf(gnuplot, "set xlabel \"1 BIN = %6.3f %s\"\n", ACQ->Bin2EUnit[Channel], ACQ->EUnitName);
            else
                fprintf(gnuplot, "set xlabel \"%s\"\n", ACQ->EUnitName);
            plotfile = fopen("Ehisto.txt", "w");
            for(i=0; i<(1<<NBIT); i++) {
                if (ACQ->EUnitMode == 0)
                    fprintf(plotfile, "%ld\n", ACQ->RebinHisto[i]); 
                else
                    fprintf(plotfile, "%6.2f %ld\n", (float)i * ACQ->Bin2EUnit[Channel] + ACQ->EUnitOffset[Channel], ACQ->RebinHisto[i]); 
                HistoEntries += ACQ->RebinHisto[i];
            }

        } else {
            plotfile = fopen("Thisto.txt", "w");
            for(i=0; i<ACQ->Thnbin; i++) {
                fprintf(plotfile, "%ld\n", ACQ->Thisto[Channel][i]); 
                HistoEntries += ACQ->Thisto[Channel][i];
            }
        }
        fclose(plotfile);
        if (ACQ->RunMode > 1)
            fprintf(gnuplot, "set title \"Channel %d - # Entries = %d\"\n", Channel, HistoEntries);
        else
            fprintf(gnuplot, "set title \"Emulation Mode - # Entries = %d\"\n", HistoEntries);
        if ((ACQ->PlotMode == 1) && (ACQ->EUnitMode == 0))
            fprintf(gnuplot, "plot 'Ehisto.txt' title 'Energy Spectrum' with histogram 1\n"); 
        else if ((ACQ->PlotMode == 1) && (ACQ->EUnitMode == 1))
            fprintf(gnuplot, "plot 'Ehisto.txt' title 'Energy Spectrum' with line 1\n"); 
        else
            fprintf(gnuplot, "plot 'Thisto.txt' title 'Time Spectrum' with histogram 2\n"); 
        fflush(gnuplot);
    }
}

/**************************************************************************//**
* \fn      void ClosePlotter(FILE *gnuplot)
* \brief   Close the plotter (gnuplot pipe)
* \param   [IN]  gnuplot: gnuplot pipe
******************************************************************************/
void ClosePlotter(FILE *gnuplot)
{
    if (gnuplot != NULL)
        pclose(gnuplot);
}

